#
#   Copyright 2022 STFC, United Kingdom Research and Innovation
#
#   Author 2022 Samuel Jones
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from trame.app import get_server
from trame.widgets import vuetify

from ccpi.viewer.CILViewer2D import CILViewer2D, SLICE_ORIENTATION_XY
from ccpi.web_viewer.trame_viewer import TrameViewer

server = get_server()
state, ctrl = server.state, server.controller


class TrameViewer2D(TrameViewer):

    def __init__(self, list_of_files: list = None):
        self.first_load = True
        super().__init__(list_of_files=list_of_files, viewer=CILViewer2D)

        self.model_choice = None
        self.background_choice = None
        self.slice_slider = None
        self.orientation_radio_buttons = None
        self.auto_window_level_button = None
        self.toggle_window_details_button = None
        self.slice_window_range_slider = None
        self.slice_window_slider = None
        self.slice_level_slider = None
        self.tracing_switch = None
        self.line_profile_switch = None
        self.interpolation_of_slice_switch = None
        self.remove_roi_button = None
        self.reset_defaults_button = None
        self.slice_interaction_col = None
        self.slice_interaction_row = None
        self.slice_interaction_section = None

        self.create_drawer_ui_elements()

        self.construct_drawer_layout()

        self.layout.content.children = [
            vuetify.VContainer(fluid=True, classes="pa-0 fill-height", children=[self.html_view])
        ]
        self.reset_defaults()
        self.layout.flush_content()

    def create_drawer_ui_elements(self):
        self.model_choice = self.create_model_selector()
        self.background_choice = self.create_background_selector()
        self.slice_slider = self.create_slice_slider()
        self.orientation_radio_buttons = self.create_orientation_radio_buttons()
        self.toggle_window_details_button = self.create_toggle_window_details_button()
        self.auto_window_level_button = self.create_auto_window_level_button()
        self.slice_window_range_slider = self.construct_slice_window_range_slider()
        self.slice_window_slider = self.construct_slice_window_slider()
        self.slice_level_slider = self.construct_slice_level_slider()
        self.tracing_switch = self.create_tracing_switch()
        self.interpolation_of_slice_switch = self.create_interpolation_of_slice_switch()
        self.remove_roi_button = self.create_remove_roi_button()
        self.reset_defaults_button = self.create_reset_defaults_button()

    def construct_drawer_layout(self):
        # The difference is that we use range slider instead of detailed sliders
        self.slice_interaction_col = vuetify.VCol([
            self.slice_slider, self.orientation_radio_buttons, self.auto_window_level_button,
            self.toggle_window_details_button, self.slice_window_range_slider, self.slice_window_slider,
            self.slice_level_slider, self.tracing_switch, self.interpolation_of_slice_switch
        ])
        self.slice_interaction_row = vuetify.VRow(self.slice_interaction_col)
        self.slice_interaction_section = vuetify.VContainer(self.slice_interaction_row)
        self.layout.drawer.children = [
            "Choose model to load", self.model_choice,
            vuetify.VDivider(), "Choose background color", self.background_choice,
            vuetify.VDivider(), self.slice_interaction_section,
            vuetify.VDivider(),
            "Use Ctrl + Click on the slice, to show the ROI of the current slice, Click and drag to resize and reposition.\n"
            "Move the ROI by using the middle mouse button.",
            vuetify.VDivider(), self.remove_roi_button,
            vuetify.VDivider(), self.reset_defaults_button
        ]

    def load_file(self, file_name, _="scalar"):
        super().load_file(file_name, windowing_method="scalar")
        if not self.first_load:
            self.update_slice_slider_data()
            self.update_slice_windowing_defaults()
            self.create_drawer_ui_elements()
            self.construct_drawer_layout()
            self.reset_defaults()
        else:
            self.first_load = False

    def update_slice_windowing_defaults(self):
        self.update_slice_data()

        if hasattr(self, "slice_window_range_slider") and self.slice_window_range_slider:
            self.slice_window_range_slider = self.construct_slice_window_range_slider()
            self.slice_level_slider = self.construct_slice_level_slider()
            self.slice_window_slider = self.construct_slice_window_slider()
            self.construct_drawer_layout()
            self.layout.flush_content()

    def create_remove_roi_button(self):
        return vuetify.VBtn("Remove ROI", hide_details=True, dense=True, solo=True, click=self.remove_roi)

    def create_tracing_switch(self):
        return vuetify.VSwitch(label="Toggle Tracing",
                               v_model=("toggle_tracing", False),
                               hide_details=True,
                               dense=True,
                               solo=True)

    def create_interpolation_of_slice_switch(self):
        return vuetify.VSwitch(
            label="Toggle Interpolation",
            v_model=("toggle_interpolation", False),
            hide_details=True,
            dense=True,
            solo=True,
        )

    def create_reset_defaults_button(self):
        return vuetify.VBtn("Reset Defaults", hide_details=True, dense=True, solo=True, click=self.reset_defaults)

    def change_tracing(self, tracing: bool):
        if tracing:
            self.cil_viewer.imageTracer.On()
        else:
            self.cil_viewer.imageTracer.Off()

    def change_interpolation(self, interpolation: bool):
        if not interpolation:
            self.cil_viewer.imageSlice.GetProperty().SetInterpolationTypeToNearest()
        else:
            self.cil_viewer.imageSlice.GetProperty().SetInterpolationTypeToLinear()
        self.cil_viewer.updatePipeline()

    def change_window_level_detail_sliders(self, show_detailed: bool):
        super().change_window_level_detail_sliders(show_detailed)

        # Reconstruct the detailed sliders
        self.slice_window_range_slider = self.construct_slice_window_range_slider()
        self.slice_window_slider = self.construct_slice_window_slider()
        self.slice_level_slider = self.construct_slice_level_slider()

        # Reconstruct the drawer and push it
        self.construct_drawer_layout()
        self.layout.flush_content()

    def remove_roi(self):
        self.cil_viewer.style.RemoveROIWidget()

    def reset_defaults(self):
        state["background_color"] = "cil_viewer_blue"
        state["slice"] = self.default_slice
        state["orientation"] = f"{SLICE_ORIENTATION_XY}"

        # resets to window-level based on 5th, 95th percentiles over volume:
        min, max = self.cil_viewer.getImageMapRange((5., 95.), "scalar")
        window, level = self.cil_viewer.getSliceWindowLevelFromRange(min, max)
        if not self.window_level_sliders_are_percentages:
            state["slice_window_range"] = (min, max)
            state["slice_window"] = window
            state["slice_level"] = level
        else:
            state["slice_window_percentiles"] = (5., 95.)
            state["slice_window_as_percentage"] = self.convert_value_to_percentage(window)
            state["slice_level_as_percentage"] = self.convert_value_to_percentage(level)

        state["toggle_tracing"] = False
        state["toggle_interpolation"] = False
        self.cil_viewer.updatePipeline()
        self.html_view.update()
        self.layout.flush_content()

    def change_slice_number(self, slice_number):
        self.cil_viewer.setActiveSlice(slice_number)
        self.cil_viewer.updatePipeline()
        self.html_view.update()



import vtk
from ccpi.viewer import SLICE_ORIENTATION_XY, SLICE_ORIENTATION_XZ, SLICE_ORIENTATION_YZ
from eqt.ui.SessionDialogs import WarningDialog


class cilviewerBoxWidget():

    @staticmethod
    def CreateFixed(viewer, outline_colour=(0, 1, 0)):
        '''
        Creates a vtkBoxWidget which the user can't move
        
        Parameters
        ----------
        viewer
            The viewer that the box widget will later be working on.
            The interactor for the widget will be set to this viewer's
            interactor.
        outline_color
            The color of the outline of the box widget

        Returns
        -------
        widget: vtk.vtkBoxWidget
        '''
        widget = vtk.vtkBoxWidget()
        widget.SetInteractor(viewer.iren)
        widget.HandlesOff()
        widget.TranslationEnabledOff()
        widget.RotationEnabledOff()
        widget.GetOutlineProperty().SetColor(outline_colour)
        widget.OutlineCursorWiresOff()
        widget.SetPlaceFactor(1)
        widget.KeyPressActivationOff()
        widget.ScalingEnabledOff()
        return widget

    @staticmethod
    def CreateMoveable(viewer, outline_colour=(0, 1, 0)):
        '''
        Creates a vtkBoxWidget which the user can move

        Parameters
        ----------
        viewer
            The viewer that the box widget will later be working on.
            The interactor for the widget will be set to this viewer's
            interactor.
        outline_color
            The color of the outline of the box widget

        Returns
        -------
        widget: vtk.vtkBoxWidget
        '''
        widget = vtk.vtkBoxWidget()
        widget.SetInteractor(viewer.iren)
        widget.HandlesOn()
        widget.TranslationEnabledOn()
        widget.RotationEnabledOff()
        widget.GetOutlineProperty().SetColor(outline_colour)
        widget.OutlineCursorWiresOff()
        widget.SetPlaceFactor(1)
        widget.KeyPressActivationOff()
        return widget

    @staticmethod
    def CreateAroundSliceOnXYPlane(viewer,
                                   axis='x',
                                   coord=0,
                                   outline_color=(1., 0., 0.),
                                   width=1,
                                   widget_name='slice_outline_widget'):
        '''
        Creates a border in [viewer], around a slice with number [coord] on
        the [axis] axis. This border is a vtkBoxWidget.
        This appears as an overlay, so appears on every projection.
        E.g. if orientation = 'y', coord = 5, width=1 then it creates a
        border around the voxels with y value of 5.
        width controls the width of the box, but it always starts at [coord].
        E.g. if orientation = 'y', coord = 5, width = 5 then it creates a
        border around the voxels with y values of 5-9.

        Note, setting width=0 creates a line rather than a box

        Parameters
        ----------
        viewer: viewer2D or viewer3D
            A viewer upon which the border will be displayed
        axis: {'x', 'y'}
            Which axis the slice is on
        coord: int
            The coordinate of the slice in world coords
        outline_color: tuple
            colour of the border
        width: float
            The width of the border. The border starts from the bound of the
            slice with number coord.
        widget_name: str
            The name to associate the border widget with in the viewer.
        '''

        widget = cilviewerBoxWidget.CreateFixed(viewer, outline_color)
        coords = cilviewerBoxWidget.GetCoordsForBoxWidgetAroundSlice(viewer, axis, coord, width)
        widget.PlaceWidget(coords)
        viewer.addWidgetReference(widget, widget_name)
        return widget

    @staticmethod
    def GetCoordsForBoxWidgetAroundSlice(viewer, axis='x', coord=0, width=1):
        '''
        Generate coordinates for positioning BoxWidget around slice
        
        Parameters
        ----------
        viewer: viewer2D or viewer3D
            A viewer upon which the border will be displayed
        orientation: {'x', 'y'}
            Which axis the slice is on
        coord: int
            The coordinate of the slice in world coords
        width: float
            The width of the border. The border starts from the bound of the
            slice with number coord.
        
        '''

        # Only makes sense to do this on the Z axis:
        render_orientation = SLICE_ORIENTATION_XY
        try:
            viewer.setSliceOrientation(render_orientation)
        except:  # method doesn't exist on 3D viewer
            pass

        spacing = viewer.img3D.GetSpacing()

        # Get maximum extents of the image in world coords
        data_extent = viewer.style.GetDataExtentInWorld()
        voxel_max_world = viewer.style.GetMinMaxVoxelsFromExtent(data_extent)[1]

        # Set the minimum world value
        world_image_min = (0, 0, 0)

        z_coords = [world_image_min[render_orientation], voxel_max_world[render_orientation]]

        if axis == 'x':
            # Displaying slice at fixed coord in X direction:
            x_coords = [coord, coord + width * spacing[0]]
            y_coords = [world_image_min[1], voxel_max_world[1]]

        else:
            # Displaying slice at fixed coord in Y direction:
            y_coords = [coord, coord + width * spacing[1]]
            x_coords = [world_image_min[0], voxel_max_world[0]]

        coords = x_coords + y_coords + z_coords

        return coords

    @staticmethod
    def CreateMoveableAtEventPosition(viewer,
                                      position,
                                      widget_name,
                                      outline_colour=(0, 1, 0),
                                      box_bounds=None,
                                      scale_factor=None):
        ''' 
        Place a moveable box widget on the viewer at the event position.
        Parameters
        ----------
        viewer
            The 2D viewer that the box will be displayed on
        position
            The event position
        widget_name
            The name the reference to the widget will be stored as
        outline_color
            The outline color of the box widget
        scale_factor
            Factor for scaling the size of the box
        '''
        # ROI Widget
        widget = cilviewerBoxWidget.CreateMoveable(viewer, outline_colour)
        if box_bounds == None:
            if scale_factor == None:
                box_bounds = cilviewerBoxWidget.GetBoxBoundsFromEventPosition(viewer, position)
            else:
                box_bounds = cilviewerBoxWidget.GetBoxBoundsFromEventPosition(viewer,
                                                                              position,
                                                                              scale_factor=scale_factor)
        # Set widget placement and make visible
        widget.PlaceWidget(box_bounds)
        widget.On()
        viewer.addWidgetReference(widget, widget_name)
        return widget

    @staticmethod
    def GetBoxBoundsFromEventPosition(viewer, position, scale_factor=0.3):
        ''' 
        Given a position of the click, translates it into world coordinates.
        Gets the current render orientation.
        Gets the extent of the data in world coordinates and its corresponding voxels. 
        Creates a warning dialog.
        Creates a 3D box around the clicked point, where the clicked point is the voxel min (lower value in all axes).
        Returns its extent in world coordinates, i.e., "box bounds". 
        Opens a warning dialog when the mouse click is outside the image.

        Parameters
        ----------
        viewer
            The 2D viewer where a projection of the 3D box will be displayed as a rectangle.
        position
            The event position in display coordinates, i.e., 3D coordinates of a mouse click on the image. 
        scale_factor
            Factor for scaling the size of the box to be smaller than the image shape.
        '''
        box_extent_world = [0, 0, 0, 0, 0, 0]

        orientation = viewer.style.GetSliceOrientation()

        coord = vtk.vtkCoordinate()
        coord.SetCoordinateSystemToDisplay()
        coord.SetValue(position[0], position[1])
        mouse_pos_world = coord.GetComputedWorldValue(viewer.style.GetRenderer())

        data_extent_world = viewer.style.GetDataExtentInWorld()
        voxel_min_world, voxel_max_world = viewer.style.GetMinMaxVoxelsFromExtent(data_extent_world)

        dialog = WarningDialog(None, message="Click inside the image.", window_title="Viewer Warning")

        box_voxel_min = [0, 0, 0]
        box_voxel_max = [0, 0, 0]
        i = [orientation, (orientation + 1) % 3, (orientation + 2) % 3]
        if voxel_min_world[i[1]] <= mouse_pos_world[i[1]] <= voxel_max_world[i[1]] and voxel_min_world[
                i[2]] <= mouse_pos_world[i[2]] <= voxel_max_world[i[2]]:

            box_voxel_min[i[0]] = voxel_min_world[i[0]]
            box_voxel_min[i[1]] = mouse_pos_world[i[1]]
            box_voxel_min[i[2]] = mouse_pos_world[i[2]]

            box_voxel_max[i[0]] = voxel_max_world[i[0]]
            box_voxel_max[i[1]] = cilviewerBoxWidget.GetTruncatedBoxCoord(box_voxel_min[i[1]], data_extent_world, i[1])
            box_voxel_max[i[2]] = cilviewerBoxWidget.GetTruncatedBoxCoord(box_voxel_min[i[2]], data_extent_world, i[2])

            box_extent_world = viewer.style.GetExtentFromVoxels(box_voxel_min, box_voxel_max)

        else:
            dialog.exec()

        return box_extent_world

    @staticmethod
    def GetTruncatedBoxCoord(start_pos, world_extent, axis_int, scale_factor=0.3):
        """
        Returns a coordinate for the edge of the box on axis [axis], scaled
        by factor [scale_factor], and truncated to make sure the box does not
        extend over the edge of the image shown.

        Parameters
        ----------
        start_pos: float
            Lower left corner value on specified axis
        world_extent:
            Array containing the extent of the world
        axis_int: int 0, 1, 2
            The axis to truncate on.
            SLICE_ORIENTATION_XY = 2  # Z
            SLICE_ORIENTATION_XZ = 1  # Y
            SLICE_ORIENTATION_YZ = 0  # X
        scale_factor: float
            The factor used to scale the length of the box, compared to the extent
            of the world on the given axis. 
            A scale factor of 1 would mean the length of the box on the given axis is 
            the same as the length of the image on that axis, unless the start position
            is greater than the minimum value on that axis, in which case, the box would
            be suitably truncated.
        """

        # get index for axis
        world_length = world_extent[2 * axis_int + 1] - world_extent[2 * axis_int]
        dist = world_length * scale_factor
        value = start_pos + dist

        # Check to make sure that it is within the image world.
        if value > world_extent[2 * axis_int + 1]:
            return world_extent[2 * axis_int + 1]
        else:
            return value


class cilviewerLineWidget():

    @staticmethod
    def CreateAtCoordOnXYPlane(viewer, axis='x', coord=0, outline_color=(1., 0., 0.), widget_name='slice_line_widget'):
        '''
            Creates a line in [viewer], at [coord] on the
            the [axis] axis. This line is made using a vtkBoxWidget.
            When created, the direction of view is forced to be SLICE_ORIENTATION_XY,
            so only the X or Y axis can be chosen.
            This appears as an overlay, so appears on every projection.
            E.g. if orientation = 'y', coord = 5, then it creates a
            line with y value of 5.

            Parameters
            ----------
            viewer: viewer2D or viewer3D
                A viewer upon which the border will be displayed
            axis: {'x', 'y'}
                Which axis the slice is on
            coord: int
                The coordinate of the slice in world coords
            outline_color: tuple
                colour of the border
            widget_name: str
                The name to associate the border widget with in the viewer.
            '''

        widget = cilviewerBoxWidget.CreateAroundSliceOnXYPlane(viewer,
                                                               axis,
                                                               coord,
                                                               outline_color,
                                                               width=0,
                                                               widget_name=widget_name)
        return widget



import os
import unittest

import h5py
import numpy as np
import vtk
from ccpi.viewer.utils.conversion import Converter, calculate_target_downsample_shape
from ccpi.viewer.utils.io import ImageReader, cilviewerHDF5Writer, cilviewerHDF5Reader


class TestImageReaderAndWriter(unittest.TestCase):

    def setUp(self):
        # Generate random 3D array:
        bits = 16
        self.bytes_per_element = int(bits / 8)
        self.input_3D_array = np.random.randint(10, size=(5, 10, 6), dtype=eval(f"np.uint{bits}"))
        # write to HDF5: -----------
        self.hdf5_filename_3D = 'test_3D_data.h5'
        with h5py.File(self.hdf5_filename_3D, 'w') as f:
            f.create_dataset('ImageData', data=self.input_3D_array)
        # write to NUMPY: ----------
        self.numpy_filename_3D = 'test_3D_data.npy'
        np.save(self.numpy_filename_3D, self.input_3D_array)
        # write to MHA: --------------
        self.mha_filename_3D = 'test_3D_data_1.mha'
        vtkimage = Converter.numpy2vtkImage(self.input_3D_array)
        writer = vtk.vtkMetaImageWriter()
        writer.SetInputData(vtkimage)
        writer.SetCompression(0)
        writer.SetFileName(self.mha_filename_3D)
        writer.Write()
        # write to raw: -------------
        bytes_3D_array = bytes(self.input_3D_array)
        self.raw_filename_3D = 'test_3D_data.raw'
        with open(self.raw_filename_3D, 'wb') as f:
            f.write(bytes_3D_array)
        self.raw_image_attrs = {
            'shape': np.shape(self.input_3D_array),
            'is_fortran': False,
            'is_big_endian': False,
            'typecode': str(self.input_3D_array.dtype)
        }

        # Create TIFF Files
        fnames = []
        arr = self.input_3D_array
        self.tiff_dirname = 'tiff_files'
        if not os.path.exists('tiff_files'):
            os.mkdir(self.tiff_dirname)
        from PIL import Image
        for i in range(arr.shape[0]):
            fname = 'tiff_files/tiff_test_file_{:03d}.tiff'.format(i)
            fnames.append(os.path.abspath(fname))
            # Using vtk the Y axis gets reversed
            # vtk_image = Converter.numpy2vtkImage(np.expand_dims(arr[i,:,:], axis=0))
            # twriter.SetFileName(fnames[-1])
            # twriter.SetInputData(vtk_image)
            # twriter.Write()
            im = Image.fromarray(arr[i])
            im.save(fnames[-1])

        self.tiff_fnames = fnames

        self.vtk_image = Converter.numpy2vtkImage(self.input_3D_array)

    def _test_read_full_size_data(self, reader):
        array_image_data = reader.Read()
        read_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(self.input_3D_array, read_array)

    def _test_resampling_not_acq_data(self, reader, target_size):
        image = reader.Read()
        extent = image.GetExtent()
        resulting_shape = tuple(image.GetDimensions())
        og_shape = np.shape(self.input_3D_array)
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        og_size = og_shape[0] * og_shape[1] * og_shape[2] * self.bytes_per_element
        expected_shape = calculate_target_downsample_shape(target_size, og_size, og_shape)
        self.assertEqual(resulting_shape, expected_shape)

    def _test_resample_size_bigger_than_image_size(self, reader):
        image = reader.Read()
        extent = image.GetExtent()
        og_shape = np.shape(self.input_3D_array)
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        expected_shape = og_shape
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        self.assertEqual(resulting_shape, expected_shape)
        resulting_array = Converter.vtk2numpy(image)
        np.testing.assert_array_equal(self.input_3D_array, resulting_array)

    def _test_resampling_acq_data(self, reader, target_size):
        og_shape = np.shape(self.input_3D_array)
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        og_size = og_shape[0] * og_shape[1] * og_shape[2] * self.bytes_per_element
        image = reader.Read()
        extent = image.GetExtent()
        shape_not_acquisition = calculate_target_downsample_shape(target_size, og_size, og_shape, acq=True)
        expected_size = shape_not_acquisition[0] * \
            shape_not_acquisition[1]*shape_not_acquisition[2]
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        resulting_size = resulting_shape[0] * \
            resulting_shape[1]*resulting_shape[2]
        # angle (z direction) is first index in numpy array, and in cil
        # but it is the last in vtk.
        resulting_z_shape = extent[5] + 1
        og_z_shape = np.shape(self.input_3D_array)[0]
        self.assertEqual(resulting_size, expected_size)
        self.assertEqual(resulting_z_shape, og_z_shape)

    def test_read(self):
        '''Test reading each format without resampling or cropping'''
        readers = {}
        readers['hdf5'] = ImageReader(file_name=self.hdf5_filename_3D, resample=False, hdf5_dataset_name="ImageData")
        readers['numpy'] = ImageReader(file_name=self.numpy_filename_3D, resample=False)
        readers['mha'] = ImageReader(file_name=self.mha_filename_3D, resample=False)
        readers['raw'] = ImageReader(file_name=self.raw_filename_3D,
                                     resample=False,
                                     raw_image_attrs=self.raw_image_attrs)
        readers['tiff_file'] = ImageReader(file_name=self.tiff_fnames, resample=False)
        readers['tiff_dir'] = ImageReader(file_name=self.tiff_dirname, resample=False)
        readers['vtk'] = ImageReader(vtk_image=self.vtk_image, resample=False)

        for i, reader in enumerate(readers.values()):
            with self.subTest(reader_name=list(readers.keys())[i]):
                self._test_read_full_size_data(reader)

    def test_read_resample(self):

        og_shape = np.shape(self.input_3D_array)
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        og_size = og_shape[0] * og_shape[1] * og_shape[2]

        # Tests image with correct target size is generated by resample reader:
        target_size = 100
        readerhdf5 = ImageReader(file_name=self.hdf5_filename_3D,
                                 target_size=target_size,
                                 resample_z=True,
                                 hdf5_dataset_name="ImageData")
        self._test_resampling_not_acq_data(readerhdf5, target_size)
        readernpy = ImageReader(file_name=self.numpy_filename_3D, target_size=target_size, resample_z=True)
        self._test_resampling_not_acq_data(readernpy, target_size)
        readermhd = ImageReader(file_name=self.mha_filename_3D, target_size=target_size, resample_z=True)
        self._test_resampling_not_acq_data(readermhd, target_size)
        readerraw = ImageReader(file_name=self.raw_filename_3D,
                                target_size=target_size,
                                resample_z=True,
                                raw_image_attrs=self.raw_image_attrs)
        self._test_resampling_not_acq_data(readerraw, target_size)
        readertiff = ImageReader(file_name=self.tiff_dirname, target_size=target_size, resample_z=True)
        self._test_resampling_not_acq_data(readertiff, target_size)
        readervtk = ImageReader(vtk_image=self.vtk_image, target_size=target_size, resample_z=True)
        self._test_resampling_not_acq_data(readervtk, target_size)

        # Now test if we get the correct z extent if we set that we
        # have acquisition data
        readerhdf5 = ImageReader(file_name=self.hdf5_filename_3D,
                                 target_size=target_size,
                                 resample_z=False,
                                 hdf5_dataset_name="ImageData")
        self._test_resampling_acq_data(readerhdf5, target_size)
        readernpy = ImageReader(file_name=self.numpy_filename_3D, target_size=target_size, resample_z=False)
        self._test_resampling_acq_data(readernpy, target_size)
        readermhd = ImageReader(file_name=self.mha_filename_3D, target_size=target_size, resample_z=False)
        self._test_resampling_acq_data(readermhd, target_size)
        readerraw = ImageReader(file_name=self.raw_filename_3D,
                                target_size=target_size,
                                resample_z=False,
                                raw_image_attrs=self.raw_image_attrs)
        self._test_resampling_acq_data(readerraw, target_size)
        readertiff = ImageReader(file_name=self.tiff_fnames, target_size=target_size, resample_z=False)
        self._test_resampling_acq_data(readertiff, target_size)
        readervtk = ImageReader(vtk_image=self.vtk_image, target_size=target_size, resample_z=False)
        self._test_resampling_acq_data(readervtk, target_size)

        # # Now test if we get the full image extent if our
        # # target size is larger than the size of the image:
        target_size = og_size * 2
        readerhdf5 = ImageReader(file_name=self.hdf5_filename_3D,
                                 target_size=target_size,
                                 hdf5_dataset_name="ImageData")
        self._test_resample_size_bigger_than_image_size(readerhdf5)
        readernpy = ImageReader(file_name=self.numpy_filename_3D, target_size=target_size)
        self._test_resample_size_bigger_than_image_size(readernpy)
        readermhd = ImageReader(file_name=self.mha_filename_3D, target_size=target_size)
        self._test_resample_size_bigger_than_image_size(readermhd)
        readerraw = ImageReader(file_name=self.raw_filename_3D,
                                target_size=target_size,
                                raw_image_attrs=self.raw_image_attrs)
        self._test_resample_size_bigger_than_image_size(readerraw)
        readertiff = ImageReader(file_name=self.tiff_dirname, target_size=target_size)
        self._test_resample_size_bigger_than_image_size(readertiff)

        readervtk = ImageReader(vtk_image=self.vtk_image, target_size=target_size)
        self._test_resample_size_bigger_than_image_size(readervtk)

    def test_read_cropped(self):
        # Test cropping the extent of a dataset
        cropped_array = self.input_3D_array[1:2, :, :]

        # HDF5: ------------------------------------------------------------------------------------------
        reader = ImageReader(file_name=self.hdf5_filename_3D,
                             crop=True,
                             resample=False,
                             target_z_extent=[1, 1],
                             hdf5_dataset_name="ImageData")

        # NOTE: the extent is in vtk so is in fortran order, whereas
        # above we had the np array in C-order so x and y cropping swapped
        array_image_data = reader.Read()
        read_cropped_array_hdf5 = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array_hdf5)

        # NUMPY: --------------------------------------------------------------
        reader = ImageReader(file_name=self.numpy_filename_3D, crop=True, resample=False, target_z_extent=[1, 1])
        array_image_data = reader.Read()
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

        # METAIMAGE: -----------------------------------------------------------------
        reader = ImageReader(file_name=self.mha_filename_3D, crop=True, resample=False, target_z_extent=[1, 1])
        array_image_data = reader.Read()
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

        # RAW: ----------------------------------------------------------------------
        reader = ImageReader(file_name=self.raw_filename_3D,
                             crop=True,
                             resample=False,
                             target_z_extent=[1, 1],
                             raw_image_attrs=self.raw_image_attrs)
        array_image_data = reader.Read()
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

        # TIFF: ------------------------------------------------------------------------
        reader = ImageReader(file_name=self.tiff_fnames, crop=True, resample=False, target_z_extent=[1, 1])
        array_image_data = reader.Read()
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

    def test_write_read_hdf5(self):
        ''''
        This:
        Reads and resamples the dataset.
        Writes out this resampled dataset and details about the original dataset
        Reads the resampled_dataset
        Checks we read back in the same array, spacing and origin
        '''
        target_size = 100
        reader = ImageReader(file_name=self.hdf5_filename_3D,
                             target_size=target_size,
                             resample_z=True,
                             hdf5_dataset_name="ImageData")

        file_to_write = 'downsampled_image.hdf5'
        resampled_image = reader.Read()
        resampled_image_array = Converter.vtk2numpy(resampled_image)

        resampled_image_attrs = reader.GetLoadedImageAttrs()
        original_image_attrs = reader.GetOriginalImageAttrs()

        # writer = ImageWriter(file_name=file_to_write, format='hdf5', datasets=[None, resampled_image], attributes=[original_image_attrs, resampled_image_attrs])
        # writer.write()

        writer = cilviewerHDF5Writer()
        writer.SetFileName(file_to_write)
        writer.SetOriginalDataset(None, original_image_attrs)
        writer.AddChildDataset(resampled_image, resampled_image_attrs)
        writer.Write()

        reader = cilviewerHDF5Reader()
        reader.SetFileName(file_to_write)
        reader.Update()
        read_resampled_image = reader.GetOutputDataObject(0)
        read_resampled_image_array = Converter.vtk2numpy(read_resampled_image)

        np.testing.assert_array_equal(resampled_image_array, read_resampled_image_array)
        np.testing.assert_array_equal(resampled_image.GetOrigin(), read_resampled_image.GetOrigin())

        # Check the attributes are the same:
        read_resampled_image_attrs = reader.GetLoadedImageAttrs()
        for key, value in resampled_image_attrs.items():
            if key == 'spacing' or key == 'origin':
                np.testing.assert_array_equal(value, read_resampled_image_attrs[key])
            else:
                self.assertEqual(value, read_resampled_image_attrs[key])
        read_original_image_attrs = reader.GetOriginalImageAttrs()
        for key, value in original_image_attrs.items():
            if key == 'spacing' or key == 'origin' or key == 'shape':
                np.testing.assert_array_equal(value, read_original_image_attrs[key])
            else:
                self.assertEqual(value, read_original_image_attrs[key])

    def tearDown(self):
        files = [self.hdf5_filename_3D, self.numpy_filename_3D, self.mha_filename_3D, self.raw_filename_3D
                 ] + self.tiff_fnames
        for f in files:
            os.remove(f)
        os.rmdir('tiff_files')


if __name__ == '__main__':
    unittest.main()



#
#   Copyright 2022 STFC, United Kingdom Research and Innovation
#
#   Author 2022 Samuel Jones
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import getopt
import os
import sys

from trame.app import get_server

from ccpi.web_viewer.trame_viewer2D import TrameViewer2D
from ccpi.web_viewer.trame_viewer3D import TrameViewer3D

server = get_server()
state, ctrl = server.state, server.controller

TRAME_VIEWER = None
VIEWER_2D = False


def reset_viewer2d():
    set_viewer2d(False)


def set_viewer2d(new_value):
    global VIEWER_2D
    VIEWER_2D = new_value


def arg_parser():
    """
    Parse the passed arguments to the current
    :return:
    """
    help_string = "web_app.py [optional args: -h, -d] <data_files>\n" \
                  "Args:\n" \
                  "-h: Show this help and exit the program\n" \
                  "-d, --2D: Use the 2D viewer instead of the 3D viewer, the default is to just use the 3D viewer."
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hd", ["2D"])
    except getopt.GetoptError:
        print(help_string)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print(help_string)
        elif opt in ("-d", "--2D"):
            global VIEWER_2D
            VIEWER_2D = True
    return data_finder()


def data_finder():
    """
    Finds all files that are needed to be passed to the TrameViewer that in a list that is digestible, using the passed args.
    :return: list, of full file paths of data in the given directory parameter
    """
    data_files = []
    for index, arg in enumerate(sys.argv):
        if index == 0 or arg[0] == '-':
            # this is the python script in index 0 and not a passed arg
            continue
        if os.path.isfile(arg):
            data_files.append(arg)
        elif os.path.isdir(arg):
            files_in_dir = os.listdir(arg)
            for file in files_in_dir:
                data_files.append(os.path.join(arg, file))
        else:
            print(f"This arg: {arg} is not a valid file or directory. Assuming it is for trame.")
    return data_files


def main() -> int:
    """
    Create the main class and run the TrameViewer
    :return: int, exit code for the program
    """
    data_files = arg_parser()
    global TRAME_VIEWER
    if not VIEWER_2D:
        TRAME_VIEWER = TrameViewer3D(data_files)
        TRAME_VIEWER.start()
    else:
        TRAME_VIEWER = TrameViewer2D(data_files)
        TRAME_VIEWER.start()
    return 0


@state.change("slice")
def update_slice(**kwargs):
    TRAME_VIEWER.change_slice_number(kwargs["slice"])


@state.change("orientation")
def change_orientation(**kwargs):
    if "orientation" in kwargs:
        orientation = kwargs["orientation"]
        if orientation is not int:
            orientation = int(orientation)
        TRAME_VIEWER.switch_to_orientation(orientation)


@state.change("opacity")
def change_opacity_mapping(**kwargs):
    if "opacity" in kwargs:
        TRAME_VIEWER.set_opacity_mapping(kwargs["opacity"])


@state.change("file_name")
def change_model(**kwargs):
    TRAME_VIEWER.load_file(kwargs['file_name'], kwargs.get('opacity', "scalar"))


@state.change("color_map")
def change_color_map(**kwargs):
    TRAME_VIEWER.change_color_map(kwargs['color_map'])


@state.change("windowing")
def change_windowing(**kwargs):
    TRAME_VIEWER.change_windowing(kwargs["windowing"][0], kwargs["windowing"][1], windowing_method=kwargs["opacity"])


@state.change("coloring")
def change_coloring(**kwargs):
    TRAME_VIEWER.change_coloring(kwargs["coloring"][0], kwargs["coloring"][1])


@state.change("slice_window")
def change_slice_window(**kwargs):
    TRAME_VIEWER.change_slice_window(kwargs["slice_window"])


@state.change("slice_window_as_percentage")
def change_slice_window_as_percentage(**kwargs):
    TRAME_VIEWER.change_slice_window_as_percentage(kwargs["slice_window_as_percentage"])


@state.change("slice_level")
def change_slice_level(**kwargs):
    TRAME_VIEWER.change_slice_level(kwargs["slice_level"])


@state.change("slice_level_as_percentage")
def change_slice_level_as_percentage(**kwargs):
    TRAME_VIEWER.change_slice_level_as_percentage(kwargs["slice_level_as_percentage"])


@state.change("slice_window_range")
def change_slice_window_level_range(**kwargs):
    min_window = kwargs["slice_window_range"][0]
    max_window = kwargs["slice_window_range"][1]
    TRAME_VIEWER.change_slice_window_level_range(min_window, max_window)


@state.change("slice_window_percentiles")
def change_slice_window_level_percentiles(**kwargs):
    min_window = kwargs["slice_window_percentiles"][0]
    max_window = kwargs["slice_window_percentiles"][1]
    TRAME_VIEWER.change_slice_window_level_percentiles(min_window, max_window)


@state.change("slice_detailed_sliders")
def change_slice_detailed_sliders(**kwargs):
    TRAME_VIEWER.change_window_level_detail_sliders(kwargs["slice_detailed_sliders"])


@state.change("slice_visibility")
def change_slice_visibility(**kwargs):
    TRAME_VIEWER.change_slice_visibility(kwargs["slice_visibility"])


@state.change("volume_visibility")
def change_volume_visibility(**kwargs):
    TRAME_VIEWER.change_volume_visibility(kwargs["volume_visibility"])


@state.change("background_color")
def change_background_color(**kwargs):
    TRAME_VIEWER.change_background_color(kwargs["background_color"])


@state.change("toggle_clipping")
def change_clipping(**kwargs):
    TRAME_VIEWER.change_clipping(kwargs["toggle_clipping"])


@state.change("toggle_tracing")
def change_tracing(**kwargs):
    TRAME_VIEWER.change_tracing(kwargs["toggle_tracing"])


@state.change("toggle_profiling")
def change_profiling(**kwargs):
    TRAME_VIEWER.change_profiling(kwargs["toggle_profiling"])


@state.change("toggle_interpolation")
def change_interpolation(**kwargs):
    TRAME_VIEWER.change_interpolation(kwargs["toggle_interpolation"])


@state.change("show_slice_histogram")
def show_slice_histogram(**kwargs):
    TRAME_VIEWER.show_slice_histogram(kwargs["show_slice_histogram"])


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
#   Copyright 2017 - 2019 Edoardo Pasca
#   Copyright 2018 Richard Smith
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
import vtk
from ccpi.viewer import (ALT_KEY, CONTROL_KEY, SHIFT_KEY, CROSSHAIR_ACTOR, CURSOR_ACTOR, HELP_ACTOR, HISTOGRAM_ACTOR,
                         LINEPLOT_ACTOR, OVERLAY_ACTOR, SLICE_ACTOR, WIPE_ACTOR, SLICE_ORIENTATION_XY,
                         SLICE_ORIENTATION_XZ, SLICE_ORIENTATION_YZ)
from ccpi.viewer.CILViewerBase import CILViewerBase
from ccpi.viewer.utils import Converter

from ccpi.viewer.widgets import cilviewerBoxWidget, SliceSliderRepresentation, SliderCallback


class CILInteractorStyle(vtk.vtkInteractorStyle):

    def __init__(self, callback):
        self.callback = callback
        self._viewer = callback
        priority = 1.0
        self.debug = False

        self.AddObserver("MouseWheelForwardEvent", self.OnMouseWheelForward, priority)
        self.AddObserver("MouseWheelBackwardEvent", self.OnMouseWheelBackward, priority)
        self.AddObserver('KeyPressEvent', self.OnKeyPress, priority)
        self.AddObserver('KeyReleaseEvent', self.OnKeyRelease, priority)
        self.AddObserver('LeftButtonPressEvent', self.OnLeftButtonPressEvent, priority)
        self.AddObserver('RightButtonPressEvent', self.OnRightButtonPressEvent, priority)
        self.AddObserver('LeftButtonReleaseEvent', self.OnLeftButtonReleaseEvent, priority)
        self.AddObserver('RightButtonReleaseEvent', self.OnRightButtonReleaseEvent, priority)
        self.AddObserver('MouseMoveEvent', self.OnMouseMoveEvent, priority)

        self.InitialEventPosition = (0, 0)

        # Initialise difference from zoom event start point
        self.dy = 0

        self._reslicing_enabled = True
        self.htext = None

    @property
    def reslicing_enabled(self):
        return self._reslicing_enabled

    @reslicing_enabled.setter
    def reslicing_enabled(self, value):
        if isinstance(value, bool):
            self._reslicing_enabled = value

    def log(self, msg):
        if self.debug:
            print(msg)

    def SetInitialEventPosition(self, xy):
        self.InitialEventPosition = xy

    def GetInitialEventPosition(self):
        return self.InitialEventPosition

    def GetKeyCode(self):
        return self.GetInteractor().GetKeyCode()

    def SetKeyCode(self, keycode):
        self.GetInteractor().SetKeyCode(keycode)

    def GetControlKey(self):
        return self.GetInteractor().GetControlKey()

    def GetShiftKey(self):
        return self.GetInteractor().GetShiftKey()

    def GetAltKey(self):
        return self.GetInteractor().GetAltKey()

    def GetEventPosition(self):
        return self.GetInteractor().GetEventPosition()

    def GetDeltaEventPosition(self):
        x, y = self.GetInteractor().GetEventPosition()
        return (x - self.InitialEventPosition[0], y - self.InitialEventPosition[1])

    def Dolly(self, factor):
        self.callback.camera.Dolly(factor)
        self.callback.ren.ResetCameraClippingRange()

    def GetDimensions(self):
        return self._viewer.img3D.GetDimensions()

    def GetInputData(self):
        return self._viewer.img3D

    def GetSliceOrientation(self):
        return self._viewer.sliceOrientation

    def SetSliceOrientation(self, orientation):
        self._viewer.sliceOrientation = orientation

    def GetActiveSlice(self):
        return self._viewer.getActiveSlice()

    def SetActiveSlice(self, sliceno):
        self._viewer.setActiveSlice(sliceno)

    def UpdatePipeline(self, reset=False):
        self._viewer.updatePipeline(reset)

    def GetActiveCamera(self):
        return self._viewer.ren.GetActiveCamera()

    def SetActiveCamera(self, camera):
        self._viewer.ren.SetActiveCamera(camera)

    def ResetCamera(self):
        self._viewer.ren.ResetCamera()

    def FlipCameraPosition(self, flip=True):
        self._viewer.flipCameraPosition = flip

    def Render(self):
        self._viewer.renWin.Render()

    def UpdateImageSlice(self):
        self._viewer.imageSlice.Update()
        self.AdjustCamera()
        self.Render()

    def AdjustCamera(self):
        self._viewer.AdjustCamera()

    def SaveRender(self, filename):
        self._viewer.saveRender(filename)

    def GetRenderWindow(self):
        return self._viewer.renWin

    def GetRenderer(self):
        return self._viewer.ren

    def GetROIWidget(self):
        return self._viewer.ROIWidget

    def SetEventActive(self, event):
        self._viewer.event.On(event)

    def SetEventInactive(self, event):
        self._viewer.event.Off(event)

    def GetViewerEvent(self, event):
        return self._viewer.event.isActive(event)

    def SetInitialCameraPosition(self, position):
        self._viewer.InitialCameraPosition = position

    def GetInitialCameraPosition(self):
        return self._viewer.InitialCameraPosition

    def SetInitialLevel(self, level):
        self._viewer.InitialLevel = level

    def GetInitialLevel(self):
        return self._viewer.InitialLevel

    def SetInitialWindow(self, window):
        self._viewer.InitialWindow = window

    def GetInitialWindow(self):
        return self._viewer.InitialWindow

    def SetROI(self, roi):
        self._viewer.ROI = roi

    def GetROI(self):
        return self._viewer.ROI

    def GetVisualisationDownsampling(self):
        return self._viewer.visualisation_downsampling

    def SetVisualisationDownsampling(self, value):
        self._viewer.setVisualisationDownsampling(value)

    def CreateAnnotationText(self, display_type, data):
        return self._viewer.createAnnotationText(display_type, data)

    def UpdateCornerAnnotation(self, text, corner):
        self._viewer.updateCornerAnnotation(text, corner)

    def GetPicker(self):
        return self._viewer.picker

    def GetCornerAnnotation(self):
        return self._viewer.cornerAnnotation

    def UpdateROIHistogram(self):
        self._viewer.updateROIHistogram()

    def UpdateLinePlot(self, imagecoordinate, display):
        self._viewer.updateLinePlot(imagecoordinate, display)

    def GetCrosshairs(self):
        actor = self._viewer.crosshairsActor
        vert = self._viewer.vertLine
        horiz = self._viewer.horizLine

        return actor, vert, horiz

    def GetImageWorldExtent(self):
        """Deprecated. Use `GetDataExtentInWorld` and `GetMinMaxVoxelsFromExtent`."""
        return self.image2world(self.GetInputData().GetExtent()[1::2])

    def GetDataExtentInWorld(self):
        """
        Compute and return the extent of the input data in the rendered world.
        """
        data_extent_image = self.GetInputData().GetExtent()
        data_extent_world = self.Image2WorldExtent(data_extent_image)
        return data_extent_world

    def GetMinMaxVoxelsFromExtent(self, extent):
        """Given the extent of a box or image, gets the voxels corresponding to the min values in all directions
        and max values in all directions."""
        voxel_min = extent[0::2]
        voxel_max = extent[1::2]
        return voxel_min, voxel_max

    def Image2WorldExtent(self, extent_image):
        """Given the extent of a box or image, gets the voxels corresponding to the min values in all directions
        and max values in all directions. Then, converts their coordinates in the world coordinate system. 
        Returns the converted extent."""
        voxel_min_image, voxel_max_image = self.GetMinMaxVoxelsFromExtent(extent_image)
        voxel_min_world = self.image2world(voxel_min_image)
        voxel_max_world = self.image2world(voxel_max_image)
        extent_world = self.GetExtentFromVoxels(voxel_min_world, voxel_max_world)
        return extent_world

    def GetExtentFromVoxels(self, voxel_min, voxel_max):
        """Given the voxels corresponding to the min values in all directions
        and max values in all directions, calculates the extent of the box or image they enclose."""
        extent = (voxel_min[0], voxel_max[0], voxel_min[1], voxel_max[1], voxel_min[2], voxel_max[2])
        return extent

    def SetCharEvent(self, char):
        self.GetInteractor().SetKeyCode(char)
        self.OnKeyPress(self.GetInteractor(), "KeyPressEvent")

    def validateValue(self, value, axis):
        return self._viewer.validateValue(value, axis)

    def InitialiseBox(self, clickPosition):
        """
        Set the initial values for the box borders
        :param clickPosition: Display coordinates for the mouse event
        """

        box_pos = cilviewerBoxWidget.GetBoxBoundsFromEventPosition(self._viewer, clickPosition)
        # Set widget placement and make visible
        self._viewer.ROIWidget.PlaceWidget(box_pos)
        self._viewer.ROIWidget.On()
        self.UpdatePipeline()

    ############### Handle events
    def OnMouseWheelForward(self, interactor, event):
        if self.GetInputData() is None:
            return
        maxSlice = self.GetInputData().GetExtent()[self.GetSliceOrientation() * 2 + 1]
        shift = interactor.GetShiftKey()
        advance = 1
        if shift:
            advance = 10

        if (self.GetActiveSlice() + advance <= maxSlice):
            self.SetActiveSlice(self.GetActiveSlice() + advance)

            self.UpdatePipeline()
        else:
            self.log("maxSlice %d request %d" % (maxSlice, self.GetActiveSlice()))

        if self.GetViewerEvent("SHOW_LINE_PROFILE_EVENT"):
            self.DisplayLineProfile(interactor, event, True)

    def OnMouseWheelBackward(self, interactor, event):
        if self.GetInputData() is None:
            return
        minSlice = self.GetInputData().GetExtent()[self.GetSliceOrientation() * 2]
        shift = interactor.GetShiftKey()
        advance = 1
        if shift:
            advance = 10
        if (self.GetActiveSlice() - advance >= minSlice):
            self.SetActiveSlice(self.GetActiveSlice() - advance)
            self.UpdatePipeline()
        else:
            self.log("minSlice %d request %d" % (minSlice, self.GetActiveSlice()))
        if self.GetViewerEvent("SHOW_LINE_PROFILE_EVENT"):
            self.DisplayLineProfile(interactor, event, True)

    def AutoWindowLevelOnVolumeRange(self, update_slice=True):
        '''Auto-adjusts window-level for the slice, based on the 5 and 95th percentiles of the whole image volume.'''
        cmin, cmax = self._viewer.getImageMapRange((5., 95.), method="scalar")
        print("Auto range for volume: ", cmin, cmax)
        window, level = self._viewer.getSliceWindowLevelFromRange(cmin, cmax)

        self._viewer.imageSlice.GetProperty().SetColorLevel(level)
        self._viewer.imageSlice.GetProperty().SetColorWindow(window)

        if update_slice:
            self.UpdateImageSlice()

    def ChangeOrientation(self, new_slice_orientation):
        orientation = self.GetSliceOrientation()
        camera = vtk.vtkCamera()
        camera.ParallelProjectionOn()
        camera.SetFocalPoint(self.GetActiveCamera().GetFocalPoint())
        camera.SetPosition(self.GetActiveCamera().GetPosition())
        self.SetInitialCameraPosition(self.GetActiveCamera().GetPosition())
        camera.SetViewUp(self.GetActiveCamera().GetViewUp())

        if new_slice_orientation == SLICE_ORIENTATION_XY:
            # Equivalent to pressing z
            if orientation == SLICE_ORIENTATION_YZ:
                self.FlipCameraPosition(True)
                camera.Elevation(90)
            elif orientation == SLICE_ORIENTATION_XZ:
                camera.Elevation(-90)
                self.FlipCameraPosition(True)
            camera.SetViewUp(0, -1, 0)
        elif new_slice_orientation == SLICE_ORIENTATION_XZ:
            # Equivalent to pressing y
            if orientation == SLICE_ORIENTATION_XY:
                camera.Elevation(90)
                self.FlipCameraPosition(True)
            elif orientation == SLICE_ORIENTATION_YZ:
                camera.Azimuth(90)
            camera.SetViewUp(0, 0, -1)
        elif new_slice_orientation == SLICE_ORIENTATION_YZ:
            # Equivalent to pressing x
            if orientation == SLICE_ORIENTATION_XY:
                camera.Azimuth(270)
            elif orientation == SLICE_ORIENTATION_XZ:
                self.FlipCameraPosition(True)
                camera.Azimuth(90)
            camera.SetViewUp(0, 0, -1)

        self.SetActiveCamera(camera)
        self.SetSliceOrientation(new_slice_orientation)
        self.UpdatePipeline(True)

    def OnKeyPress(self, interactor, event):
        al = self._viewer.axisLabelsText
        if self.GetInputData() is None:
            return
        if self.reslicing_enabled and interactor.GetKeyCode() == "x":
            self._viewer.setAxisLabels(['', al[1], al[2]], False)
            self.ChangeOrientation(SLICE_ORIENTATION_YZ)
        elif self.reslicing_enabled and interactor.GetKeyCode() == "y":
            self._viewer.setAxisLabels([al[0], '', al[2]], False)
            self.ChangeOrientation(SLICE_ORIENTATION_XZ)
        elif self.reslicing_enabled and interactor.GetKeyCode() == "z":
            self._viewer.setAxisLabels([al[0], al[1], ''], False)
            self.ChangeOrientation(SLICE_ORIENTATION_XY)
        elif interactor.GetKeyCode() == "a":
            self._viewer.autoWindowLevelOnSliceRange()
        elif interactor.GetKeyCode() == "s":
            filename = "current_render"
            self.SaveRender(filename)

        elif interactor.GetKeyCode() == "q":
            self.log("Render loop terminating by pressing %s" % (interactor.GetKeyCode(), ))
            interactor.SetKeyCode("e")
            self.OnKeyPress(interactor, event)

        elif interactor.GetKeyCode() == "l":
            if self.GetViewerEvent("SHOW_LINE_PROFILE_EVENT"):

                self.SetEventInactive("SHOW_LINE_PROFILE_EVENT")
                self.DisplayLineProfile(interactor, event, False)
            else:
                self.SetEventActive("SHOW_LINE_PROFILE_EVENT")
                self.DisplayLineProfile(interactor, event, True)

        elif interactor.GetKeyCode() == "h":
            self.DisplayHelp()
        elif interactor.GetKeyCode() == "w":
            self.SetEventActive('UPDATE_WINDOW_LEVEL_UNDER_CURSOR')
        elif interactor.GetKeyCode() == "t":
            # tracing event is captured by widget
            if (self._viewer.imageTracer.GetEnabled()):
                self._viewer.imageTracer.Off()
            else:
                self._viewer.imageTracer.On()
        elif interactor.GetKeyCode() == "i":
            # toggle interpolation of slice actor
            is_interpolated = self._viewer.imageSlice.GetProperty().GetInterpolationType()
            if is_interpolated:
                self._viewer.imageSlice.GetProperty().SetInterpolationTypeToNearest()
            else:
                self._viewer.imageSlice.GetProperty().SetInterpolationTypeToLinear()
            self._viewer.updatePipeline()

        elif interactor.GetKeyCode() == '1':
            ev = 'RECTILINEAR_WIPE'
            if self.GetViewerEvent(ev):
                self.SetEventInactive(ev)
            # ImageWithOverlay
            self._viewer.setVisualisationToImageWithOverlay()
            self.AdjustCamera()
            self.Render()

        elif interactor.GetKeyCode() == '2':
            if self._viewer.image2 is not None:
                if self._viewer.vis_mode != CILViewer2D.RECTILINEAR_WIPE:
                    self._viewer.setVisualisationToRectilinearWipe()
                    orient = ['x', 'y', 'z']
                    self.SetCharEvent(orient[self.GetSliceOrientation()])
                    self.SetEventActive('RECTILINEAR_WIPE')

        else:
            self.log("Unhandled event %s" % (interactor.GetKeyCode()))

    def OnKeyRelease(self, interactor, event):
        # remove events on key release
        events = ['UPDATE_WINDOW_LEVEL_UNDER_CURSOR']
        for ev in events:
            if self.GetViewerEvent(ev):
                # print ("remove event {}".format(ev))
                self.SetEventInactive(ev)

    def RemoveROIWidget(self):
        self.SetEventActive("DELETE_ROI_EVENT")
        self.GetROIWidget().Off()
        self._viewer.updateCornerAnnotation("", 1, False)
        self.SetDisplayHistogram(False)
        self.Render()

    def OnLeftButtonPressEvent(self, interactor, event):
        # print ("INTERACTOR", interactor)
        if self.GetInputData() is None:
            return
        alt = interactor.GetAltKey()
        shift = interactor.GetShiftKey()
        ctrl = interactor.GetControlKey()

        self.SetInitialEventPosition(interactor.GetEventPosition())

        if ctrl and not (alt and shift):
            self.SetEventActive("CREATE_ROI_EVENT")
            position = interactor.GetEventPosition()
            self.InitialiseBox(position)
            self.SetDisplayHistogram(True)
            self.Render()
            self.log("Event %s is CREATE_ROI_EVENT" % (event))

        elif alt and not (shift and ctrl):
            self.RemoveROIWidget()
            self.log("Event %s is DELETE_ROI_EVENT" % (event))

        elif not (ctrl and alt and shift):
            self.SetEventActive("PICK_EVENT")
            self.HandlePickEvent(interactor, event)
            self.log("Event %s is PICK_EVENT" % (event))

    def SetDisplayHistogram(self, display):
        if display:
            if (self._viewer.displayHistogram == 0):
                #self.GetRenderer().AddActor(self._viewer.histogramPlotActor)
                self._viewer.AddActor(self._viewer.histogramPlotActor, HISTOGRAM_ACTOR)
                self.firstHistogram = 1
                self.Render()

            self._viewer.histogramPlotActor.VisibilityOn()
            self._viewer.displayHistogram = True
        else:
            self._viewer.histogramPlotActor.VisibilityOff()
            self._viewer.displayHistogram = False

    def OnLeftButtonReleaseEvent(self, interactor, event):
        interactor = self._viewer.getInteractor()

        if self.GetViewerEvent("CREATE_ROI_EVENT"):
            self.OnROIModifiedEvent(interactor, event)

        elif self.GetViewerEvent("PICK_EVENT"):
            self.HandlePickEvent(interactor, event)

        # Turn off CREATE_ROI and PICK_EVENT
        self.SetEventInactive("CREATE_ROI_EVENT")
        self.SetEventInactive("PICK_EVENT")
        self.SetEventInactive("DELETE_ROI_EVENT")

    def OnRightButtonPressEvent(self, interactor, event):
        if self.GetInputData() is None:
            return
        alt = interactor.GetAltKey()
        shift = interactor.GetShiftKey()
        ctrl = interactor.GetControlKey()

        self.SetInitialEventPosition(interactor.GetEventPosition())

        if alt and not (ctrl and shift):
            self.SetEventActive("WINDOW_LEVEL_EVENT")
            if self._viewer.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
                self.log("Event %s is WINDOW_LEVEL_EVENT" % (event))
                self.HandleWindowLevel(interactor, event)
        elif shift and not (ctrl and alt):
            self.SetEventActive("ZOOM_EVENT")
            self.SetInitialCameraPosition(self.GetActiveCamera().GetPosition())
            self.log("Event %s is ZOOM_EVENT" % (event))
        elif ctrl and not (shift and alt):
            self.SetEventActive("PAN_EVENT")
            self.SetInitialCameraPosition(self.GetActiveCamera().GetPosition())
            self.log("Event %s is PAN_EVENT" % (event))

    def OnRightButtonReleaseEvent(self, interactor, event):
        self.log(event)
        if self.GetViewerEvent("WINDOW_LEVEL_EVENT"):
            if self._viewer.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
                self.SetInitialLevel(self._viewer.imageSlice.GetProperty().GetColorLevel())
                self.SetInitialWindow(self._viewer.imageSlice.GetProperty().GetColorWindow())
        elif self.GetViewerEvent("ZOOM_EVENT") or self.GetViewerEvent("PAN_EVENT"):
            self.SetInitialCameraPosition(())

            # Reset difference from start of zoom event
            self.dy = 0

        # self.SetViewerEvent( ViewerEvent.NO_EVENT )
        self.SetEventInactive("WINDOW_LEVEL_EVENT")
        self.SetEventInactive("ZOOM_EVENT")
        self.SetEventInactive("PAN_EVENT")

    def BoxExtentCheck(self, box_extent_world):
        box_voxel_min, box_voxel_max = self.GetMinMaxVoxelsFromExtent(box_extent_world)
        box_voxel_min = list(box_voxel_min)
        box_voxel_max = list(box_voxel_max)
        # Get maximum extents of the image in world coords
        data_extent = self.GetDataExtentInWorld()
        voxel_min_world, voxel_max_world = self.GetMinMaxVoxelsFromExtent(data_extent)

        i = [self.GetSliceOrientation()]
        i.extend([(i[0] + 1) % 3, (i[0] + 2) % 3])
        if box_voxel_min[i[1]] < voxel_min_world[i[1]]:
            box_voxel_min[i[1]] = voxel_min_world[i[1]]
        if box_voxel_min[i[2]] < voxel_min_world[i[2]]:
            box_voxel_min[i[2]] = voxel_min_world[i[2]]
        if box_voxel_max[i[1]] > voxel_max_world[i[1]]:
            box_voxel_max[i[1]] = voxel_max_world[i[1]]
        if box_voxel_max[i[2]] > voxel_max_world[i[2]]:
            box_voxel_max[i[2]] = voxel_max_world[i[2]]
        box_extent_world = self.GetExtentFromVoxels(box_voxel_min, box_voxel_max)
        return box_extent_world

    def GetBoxWidgetExtentInWorld(self, box_widget):
        ''' Returns the extent of a box_widget (vtkBoxWidget)
        which is present on the viewer, in the image coordinate system.'''
        pd = vtk.vtkPolyData()
        box_widget.GetPolyData(pd)
        box_extent_world = pd.GetBounds()
        return box_extent_world

    def GetBoxWidgetExtentInImage(self, box_widget):
        ''' Returns the extent of a box_widget (vtkBoxWidget)
        which is present on the viewer, in the image coordinate system.'''
        pd = vtk.vtkPolyData()
        box_widget.GetPolyData(pd)
        box_extent_world = pd.GetBounds()
        box_voxel_min_world, box_voxel_max_world = self.GetMinMaxVoxelsFromExtent(box_extent_world)
        box_voxel_min_image = self.createVox(box_voxel_min_world)
        box_voxel_max_image = self.createVox(box_voxel_max_world)
        box_extent_image = self.GetExtentFromVoxels(box_voxel_min_image, box_voxel_max_image)
        return box_extent_image

    def OnROIModifiedEvent(self, interactor, event):
        # Get bounds from 3D ROI
        pd = vtk.vtkPolyData()
        self.GetROIWidget().GetPolyData(pd)
        bounds = pd.GetBounds()

        # Get maximum extents of the image in world coords
        bounds = self.BoxExtentCheck(bounds)
        self._viewer.ROIWidget.PlaceWidget(bounds)

        data_extent = self.GetDataExtentInWorld()
        #self._viewer.ROIWidget.On()
        #self.UpdatePipeline()
        voxel_min_world, voxel_max_world = self.GetMinMaxVoxelsFromExtent(data_extent)
        vox1 = self.createVox(voxel_min_world)
        vox2 = self.createVox(voxel_max_world)

        # Set the ROI using image coordinates
        self.SetROI((vox1, vox2))
        roi = self.GetROI()

        # Debug messages
        self.log("ROI {0}".format(roi))
        self.log("Pixel1 %d,%d,%d Value %f" % vox1)
        self.log("Pixel2 %d,%d,%d Value %f" % vox2)

        # Calculate the size of the ROI
        if self.GetSliceOrientation() == SLICE_ORIENTATION_XY:
            self.log("slice orientation : XY")
            x = abs(roi[1][0] - roi[0][0])
            y = abs(roi[1][1] - roi[0][1])
            z = abs(roi[1][2] - roi[0][2])

        elif self.GetSliceOrientation() == SLICE_ORIENTATION_XZ:
            self.log("slice orientation : XZ")
            x = abs(roi[1][0] - roi[0][0])
            y = abs(roi[1][2] - roi[0][2])
            z = abs(roi[1][1] - roi[0][1])

        elif self.GetSliceOrientation() == SLICE_ORIENTATION_YZ:
            self.log("slice orientation : YZ")
            x = abs(roi[1][1] - roi[0][1])
            y = abs(roi[1][2] - roi[0][2])
            z = abs(roi[1][1] - roi[0][1])

        # Update the text bottom right of the viewer and histogram
        roi_data = (x, y, z, float(x * y) / 1024.)
        text = self.CreateAnnotationText("roi", roi_data)
        self.log(text)
        self.UpdateCornerAnnotation(text, 1)
        self.UpdateROIHistogram()
        # self.SetViewerEvent( ViewerEvent.NO_EVENT )
        self.SetEventInactive("CREATE_ROI_EVENT")

    def OnTracerModifiedEvent(self, interactor, event):
        # Makes sure tracer is visible on current slice:
        self.UpdatePipeline()

###########################  Coordinate conversion methods ############################

    def display2world(self, displayCoords):
        """
        Takes display coordinates and converts them into world coordinates

        :param displayCoords: tuple containing the X,Y coordinates for the point in the display
        :return: The computed world coordinate of the given point as double (x,y,z)
        """
        coord = vtk.vtkCoordinate()
        coord.SetCoordinateSystemToDisplay()
        coord.SetValue(displayCoords[0], displayCoords[1])

        return coord.GetComputedWorldValue(self.GetRenderer())

    def world2display(self, world_coords):
        """
        Takes coordinates in the world system and converts them to 2D display coordinates
        :param world_coords: (x,y,z) coordinate in the world
        :return: (x,y) screen coordinate
        """

        vc = vtk.vtkCoordinate()
        vc.SetCoordinateSystemToWorld()
        vc.SetValue(world_coords)

        return vc.GetComputedDoubleDisplayValue(self.GetRenderer())

    def display2imageCoordinate(self, viewerposition, subvoxel=False):
        """
        Convert display coordinates into image coordinates and add the pixel value

        :param viewerposition: (x,y) position of the selected point in the display window
        :return: (x,y,z,a) x,y,z index of the selected slice + a, the pixel value
        """
        vc = vtk.vtkCoordinate()
        vc.SetCoordinateSystemToViewport()
        vc.SetValue(viewerposition[0:2] + (0.0, ))

        pickPosition = list(vc.GetComputedWorldValue(self.GetRenderer()))
        # print ("PICK POS", pickPosition)

        pickPosition[self.GetSliceOrientation()] = \
            self.GetInputData().GetSpacing()[self.GetSliceOrientation()]  * (self.GetActiveSlice()) # + self.GetInputData().GetOrigin()[self.GetSliceOrientation()])
        self.log("Pick Position " + str(pickPosition))

        if (pickPosition != [0, 0, 0]):

            imagePosition = self.world2imageCoordinate(pickPosition)
            imagePositionF = self.world2imageCoordinateFloat(pickPosition)
            extent = self._viewer.img3D.GetExtent()

            # make sure the pick is on the image
            if imagePosition[0] < extent[0]:
                imagePosition[0] = extent[0]
            if imagePosition[0] > extent[1]:
                imagePosition[0] = extent[1]
            if imagePosition[1] < extent[2]:
                imagePosition[1] = extent[2]
            if imagePosition[1] > extent[3]:
                imagePosition[1] = extent[3]
            if imagePosition[2] < extent[4]:
                imagePosition[2] = extent[4]
            if imagePosition[2] > extent[5]:
                imagePosition[2] = extent[5]

            self.log("imagePosition pre validate {}".format(imagePosition))

            pixelValue = self.GetInputData().GetScalarComponentAsDouble(imagePosition[0], imagePosition[1],
                                                                        imagePosition[2], 0)
            if self._viewer.rescale[0]:
                scale, shift = self._viewer.rescale[1]
                # pix = orig * scale + shift
                # orig = (-shift + pix) / scale
                pixelValue = (-shift + pixelValue) / scale

            if subvoxel:
                for i in range(3):
                    if not i == self.GetSliceOrientation():
                        imagePosition[i] = imagePositionF[i]

            return (self.validateValue(imagePosition[0], 'x'), self.validateValue(imagePosition[1], 'y'),
                    self.validateValue(imagePosition[2], 'z'), pixelValue)
        else:
            return (0, 0, 0, 0)

    def createVox(self, world_coordinates):

        # Translate the world coordinates to an image index
        imagePosition = self.world2imageCoordinate(world_coordinates)

        pixelValue = self.GetInputData().GetScalarComponentAsDouble(imagePosition[0], imagePosition[1],
                                                                    imagePosition[2], 0)
        if self._viewer.rescale[0]:
            scale, shift = self._viewer.rescale[1]
            pixelValue = (-shift + pixelValue) / scale
        return (self.validateValue(imagePosition[0],
                                   'x'), self.validateValue(imagePosition[1],
                                                            'y'), self.validateValue(imagePosition[2], 'z'), pixelValue)

    def world2imageCoordinate(self, world_coordinates):
        """
        Convert from the world or global coordinates to image coordinates
        :param world_coordinates: (x,y,z)
        :return: rounded to next integer (x,y,z) in image coorindates eg. slice index
        """

        dims = self.GetInputData().GetDimensions()
        self.log(dims)
        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        return [round((world_coordinates[i]) / spac[i] - orig[i]) for i in range(3)]

    def world2imageCoordinateFloat(self, world_coordinates):
        """
        Convert from the world or global coordinates to image coordinates
        :param world_coordinates: (x,y,z)
        :return: float (x,y,z) in image coorindates eg. slice index
        """
        dims = self.GetInputData().GetDimensions()
        self.log(dims)
        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        return [(world_coordinates[i]) / spac[i] - orig[i] for i in range(3)]

    def image2world(self, image_coordinates):
        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        return [(image_coordinates[i]) * spac[i] + orig[i] for i in range(3)]

    def imageCoordinate2display(self, imageposition):
        '''
        Convert image coordinates back into viewer coordinates
        :param imageposition: (x,y,z) coordinates in image coordinates
        :return: (x,y,z) coordinates for the window
        '''
        # Truncate to first 3 values x,y,z. Not interested in pixel value.
        ip = imageposition[0:3]

        spac = self.GetInputData().GetSpacing()
        orig = self.GetInputData().GetOrigin()

        # Convert image coordiantes to world coordinates
        world_coord = [spac[i] * (ip[i] - orig[i]) for i in range(3)]

        vc = vtk.vtkCoordinate()
        vc.SetCoordinateSystemToWorld()
        vc.SetValue(world_coord)

        return vc.GetComputedDoubleViewportValue(self.GetRenderer())

    def display2normalisedViewport(self, display_coords):
        wsize = self.GetRenderWindow().GetSize()

        x = display_coords[0] / wsize[0]
        y = display_coords[1] / wsize[1]

        return x, y

    def OnMouseMoveEvent(self, interactor, event):
        if self.GetInputData() is not None:
            if self.GetViewerEvent("WINDOW_LEVEL_EVENT"):
                self.log("Event %s is WINDOW_LEVEL_EVENT" % (event))
                self.HandleWindowLevel(interactor, event)

            elif self.GetViewerEvent("PICK_EVENT"):
                self.HandlePickEvent(interactor, event)

            elif self.GetViewerEvent("ZOOM_EVENT"):
                self.HandleZoomEvent(interactor, event)

            elif self.GetViewerEvent("PAN_EVENT"):
                self.HandlePanEvent(interactor, event)

            elif self.GetViewerEvent("SHOW_LINE_PROFILE_EVENT"):
                self.DisplayLineProfile(interactor, event, True)
            elif self.GetViewerEvent('UPDATE_WINDOW_LEVEL_UNDER_CURSOR'):
                x, y = interactor.GetEventPosition()
                ic = self.display2imageCoordinate((x, y))
                print(x, y, ic, "image coordinate")
                whole_extent = self._viewer.img3D.GetExtent()
                around = numpy.min(numpy.asarray([whole_extent[1], whole_extent[3], whole_extent[5]])) // 10
                print(around, "around")
                extent = [
                    ic[0] - around, ic[0] + around, ic[1] - around, ic[1] + around, ic[2] - around, ic[2] + around
                ]

                orientation = self._viewer.sliceOrientation

                extent[orientation * 2] = self.GetActiveSlice()
                extent[orientation * 2 + 1] = self.GetActiveSlice()

                if extent[0] < whole_extent[0]:
                    extent[0] = whole_extent[0]
                if extent[1] > whole_extent[1]:
                    extent[1] = whole_extent[1]
                if extent[2] < whole_extent[2]:
                    extent[2] = whole_extent[2]
                if extent[3] > whole_extent[3]:
                    extent[3] = whole_extent[3]
                if extent[4] < whole_extent[4]:
                    extent[4] = whole_extent[4]
                if extent[5] > whole_extent[5]:
                    extent[5] = whole_extent[5]
                # get mouse location

                print(*extent, "w extent")
                self._viewer.voicursor.SetInputData(self._viewer.img3D)
                self._viewer.voicursor.SetVOI(*extent)

                self._viewer.voicursor.Update()
                # set window/level for current slices

                self._viewer.iacursor.SetInputConnection(self._viewer.voicursor.GetOutputPort())
                self._viewer.iacursor.SetAutoRangePercentiles(1.0, 99.)
                self._viewer.iacursor.Update()
                # reset color/window
                cmin, cmax = self._viewer.iacursor.GetAutoRange()

                window, level = self._viewer.getSliceWindowLevelFromRange(cmin, cmax)

                self.SetInitialLevel(level)
                self.SetInitialWindow(window)

                self._viewer.imageSlice.GetProperty().SetColorLevel(self.GetInitialLevel())
                self._viewer.imageSlice.GetProperty().SetColorWindow(self.GetInitialWindow())

                self.UpdateImageSlice()

            elif self.GetViewerEvent('RECTILINEAR_WIPE'):
                # get event in image coordinate
                x, y, z, pix = self.display2imageCoordinate(interactor.GetEventPosition())
                # update the wipe depending on the slice orientation
                slice_orientation = self._viewer.getSliceOrientation()
                if slice_orientation == SLICE_ORIENTATION_XY:
                    self._viewer.wipe.SetAxis(0, 1)
                    self._viewer.wipe.SetPosition(x, y)
                elif slice_orientation == SLICE_ORIENTATION_XZ:
                    self._viewer.wipe.SetAxis(0, 2)
                    self._viewer.wipe.SetPosition(x, z)
                elif slice_orientation == SLICE_ORIENTATION_YZ:
                    self._viewer.wipe.SetAxis(2, 1)
                    self._viewer.wipe.SetPosition(z, y)

                self.UpdatePipeline()

    def DisplayHelp(self):
        help_actor = self._viewer.helpActor
        image_slice = self._viewer.imageSlice

        if help_actor.GetVisibility():
            help_actor.VisibilityOff()
            image_slice.VisibilityOn()
            self.Render()
            return

        font_size = 16

        # Create the text mappers and the associated Actor2Ds.

        # The font and text properties (except justification) are the same for
        # each multi line mapper. Let's create a common text property object
        multiLineTextProp = vtk.vtkTextProperty()
        multiLineTextProp.SetFontSize(font_size)
        multiLineTextProp.SetFontFamilyToArial()
        multiLineTextProp.BoldOn()
        multiLineTextProp.ItalicOn()
        multiLineTextProp.ShadowOn()
        multiLineTextProp.SetLineSpacing(1.3)

        # The text is on multiple lines and center-justified (both horizontal and
        # vertical).
        textMapperC = vtk.vtkTextMapper()
        if self.htext == None:
            self.htext = """
            Mouse Interactions:
                - Slice: Mouse Scroll
                - Quick Slice: Shift + Mouse Scroll
                - Pick: Left Click
                - Zoom: Shift + Right Mouse + Move Up/Down
                - Pan: Ctrl + Right Mouse + Move
                - Adjust Window: Alt+ Right Mouse + Move Up/Down
                - Adjust Level: Alt + Right Mouse + Move Left/Right

            Region of Interest (ROI):
                - Create: Ctrl + Left Click
                - Delete: Alt + Left Click
                - Resize: Click + Drag handles
                - Translate: Middle Mouse + Move within ROI

            Keyboard Interactions:
                h: This help
                x: YZ Plane
                y: XZ Plane
                z: XY Plane
                a: Whole image Auto Window/Level
                w: Region around cursor Auto Window/Level
                l: Line Profile at cursor
                s: Save Current Image
                t: Tracing
                i: Toggle interpolation of slice
                """
        textMapperC.SetInput(self.htext)
        tprop = textMapperC.GetTextProperty()
        tprop.ShallowCopy(multiLineTextProp)
        tprop.SetJustificationToLeft()
        tprop.SetVerticalJustificationToCentered()
        tprop.SetColor(0, 1, 0)

        help_actor.SetMapper(textMapperC)
        help_actor.VisibilityOn()
        image_slice.VisibilityOff()
        self.Render()

    def HandleZoomEvent(self, interactor, event):
        camera = self.GetActiveCamera()

        # Extract change from start of event
        dx, dy = interactor.GetDeltaEventPosition()
        window_y_size = self.GetRenderWindow().GetSize()[1]

        # Determine whether the user is zooming in or out
        change = dy - self.dy

        # Make sure that a change has been registered
        if change != 0:
            # >1 zoom in, <1 zoom out
            camera.Zoom(1 + change / window_y_size)
            self.Render()

        # Set the overall change value
        self.dy = dy

    def HandlePanEvent(self, interactor, event):
        #Camera uses world coordinates, not display coordinates so we have to make a coneversion
        interactor_event_position = interactor.GetEventPosition()
        interactor_initial_event_position = interactor.GetInitialEventPosition()

        event_position = interactor.image2world(interactor.display2imageCoordinate(interactor_event_position)[:-1])
        initial_event_position = interactor.image2world(
            interactor.display2imageCoordinate(interactor_initial_event_position)[:-1])

        #Update initial position to current event position, ready for next panning event:
        interactor.SetInitialEventPosition(interactor_event_position)

        change = []
        for i in range(len(event_position)):
            change.append(event_position[i] - initial_event_position[i])

        camera = self.GetActiveCamera()

        newposition = [i for i in self.GetInitialCameraPosition()]
        newfocalpoint = [i for i in camera.GetFocalPoint()]

        for i in range(len(event_position)):
            newposition[i] -= change[i]
            newfocalpoint[i] -= change[i]

        camera.SetFocalPoint(newfocalpoint)
        camera.SetPosition(newposition)
        self.SetInitialCameraPosition(newposition)

        self.Render()

    def HandleWindowLevel(self, interactor, event):
        dx, dy = interactor.GetDeltaEventPosition()
        self.log("Event delta %d %d" % (dx, dy))
        size = self.GetRenderWindow().GetSize()

        dx = 1 * dx / size[0]
        dy = 1 * dy / size[1]
        window = self.GetInitialWindow()
        level = self.GetInitialLevel()

        if abs(window) > 0.01:
            dx = dx * window
        else:
            dx = dx * (lambda x: -0.01 if x < 0 else 0.01)(window)

        if abs(level) > 0.01:
            dy = dy * level
        else:
            dy = dy * (lambda x: -0.01 if x < 0 else 0.01)(level)

        # Abs so that direction does not flip

        if window < 0.0:
            dx = -1 * dx
        if level < 0.0:
            dy = -1 * dy

        # Compute new window level
        newWindow = window + dx
        newLevel = level + dy

        # Stay away from zero and really small numbers
        if abs(newWindow) < 0.01:
            newWindow = 0.01 * (lambda x: -1 if x < 0 else 1)(newWindow)

        if abs(newLevel) < 0.01:
            newLevel = 0.01 * (lambda x: -1 if x < 0 else 1)(newLevel)

        self._viewer.imageSlice.GetProperty().SetColorLevel(newLevel)
        self._viewer.imageSlice.GetProperty().SetColorWindow(newWindow)
        self.log("new level {0} window {1}".format(newLevel, newWindow))
        self.UpdateImageSlice()

    def HandlePickEvent(self, interactor, event):
        position = interactor.GetEventPosition()
        vox = self.display2imageCoordinate(position)
        self.last_picked_voxel = vox
        # print ("Pixel %d,%d,%d Value %f" % vox )
        self._viewer.cornerAnnotation.VisibilityOn()
        text = self.CreateAnnotationText("pick", vox)
        self.UpdateCornerAnnotation(text, 0)
        self.Render()

    def DisplayLineProfile(self, interactor, event, display):
        x, y = interactor.GetEventPosition()
        ic = self.display2imageCoordinate((x, y))
        self.UpdateLinePlot(ic, display)


###############################################################################


class CILViewer2D(CILViewerBase):
    '''Simple Interactive Viewer based on VTK classes'''
    # visualisation modes
    IMAGE_WITH_OVERLAY = 0
    RECTILINEAR_WIPE = 1

    def __init__(self, dimx=600, dimy=600, ren=None, renWin=None, iren=None, debug=False, enableSliderWidget=True):
        CILViewerBase.__init__(self, dimx=dimx, dimy=dimy, ren=ren, renWin=renWin, iren=iren, debug=debug)

        self.setInteractorStyle(CILInteractorStyle(self))

        self.debug = debug
        self.style.debug = debug

        self.camera = vtk.vtkCamera()
        self.camera.ParallelProjectionOn()
        self.flipCameraPosition = False
        self.ren.SetActiveCamera(self.camera)

        self.axes_initialised = False

        #Actors
        self.iacursor = vtk.vtkImageHistogramStatistics()
        self.voicursor = vtk.vtkExtractVOI()
        #self.sliceActorNo = 0

        # input 2
        self.image2 = None
        self.voi2 = vtk.vtkExtractVOI()
        self.imageSlice2 = vtk.vtkImageSlice()
        self.imageSliceMapper2 = vtk.vtkImageSliceMapper()
        self.imageSlice2.SetMapper(self.imageSliceMapper2)

        # Help text
        self.AddActor(self.helpActor, HELP_ACTOR)

        # ROI Widget
        self.ROIWidget = cilviewerBoxWidget.CreateMoveable(self)

        self.ROIWidget.AddObserver(vtk.vtkWidgetEvent.Select, self.style.OnROIModifiedEvent, 1.0)

        # edge points of the ROI
        self.ROI = ()

        # Edge points of Volumetric ROI
        self.ROIV = None

        #picker
        self.picker = vtk.vtkPropPicker()
        self.picker.PickFromListOn()
        self.picker.AddPickList(self.imageSlice)

        self.iren.SetPicker(self.picker)

        # corner annotation
        self.cornerAnnotation = vtk.vtkCornerAnnotation()
        self.cornerAnnotation.SetMaximumFontSize(12)
        self.cornerAnnotation.PickableOff()
        self.cornerAnnotation.VisibilityOff()
        self.cornerAnnotation.GetTextProperty().ShadowOn()
        self.cornerAnnotation.SetLayerNumber(1)

        #used to scale corner annotation and decide whether to interpolate:
        self.image_is_downsampled = False
        self.visualisation_downsampling = [1, 1, 1]
        self.display_unsampled_coords = False

        # cursor doesn't show up
        self.cursor = vtk.vtkCursor2D()
        self.cursorMapper = vtk.vtkPolyDataMapper2D()
        self.cursorActor = vtk.vtkActor2D()
        self.cursor.SetModelBounds(-10, 10, -10, 10, 0, 0)
        self.cursor.SetFocalPoint(0, 0, 0)
        self.cursor.AllOff()
        self.cursor.AxesOn()
        self.cursorActor.PickableOff()
        self.cursorActor.VisibilityOn()
        self.cursorActor.GetProperty().SetColor(1, 1, 1)
        self.cursorActor.SetLayerNumber(0)
        self.cursorMapper.SetInputData(self.cursor.GetOutput())
        self.cursorActor.SetMapper(self.cursorMapper)
        # self.getRenderer().AddActor(self.cursorActor)
        self.AddActor(self.cursorActor, CURSOR_ACTOR)

        # Zoom
        self.InitialCameraPosition = ()

        # XY Plot actor for histogram
        self.displayHistogram = False
        self.firstHistogram = 0
        self.roiIA = vtk.vtkImageAccumulate()
        self.roiVOI = vtk.vtkExtractVOI()
        self.histogramPlotActor.SetPosition2(0.6, 0.6)
        self.histogramPlotActor.SetPosition(0.4, 0.4)

        # XY Plot for X and Y slices
        self.displayLinePlot = False
        self.linePlot = 0
        self.lineVOIX = vtk.vtkExtractVOI()
        self.lineVOIY = vtk.vtkExtractVOI()
        self.linePlotActor = vtk.vtkXYPlotActor()
        self.linePlotActor.ExchangeAxesOff()
        self.linePlotActor.SetXTitle("")
        self.linePlotActor.SetYTitle("")
        self.linePlotActor.SetXLabelFormat("%.1f")
        self.linePlotActor.SetYLabelFormat("%.1e")
        #self.linePlotActor.SetAdjustXLabels(3)
        #self.linePlotActor.SetXTitle( "Level" )
        #self.linePlotActor.SetYTitle( "N" )
        self.linePlotActor.SetXValuesToValue()
        self.linePlotActor.SetPlotColor(0, (1, 0, 0.5))
        self.linePlotActor.SetPlotColor(1, (1, 1, 0))
        self.linePlotActor.SetPosition(0, 0.1)
        self.linePlotActor.SetPosition2(1, 0.4)

        # Makes sure that x axis only goes as far as the number of pixels in the image
        self.linePlotActor.SetAdjustXLabels(0)

        # Add legend
        self.linePlotActor.SetPlotLabel(0, 'horiz')
        self.linePlotActor.SetPlotLabel(1, 'vert')
        self.linePlotActor.LegendOn()

        # crosshair lines for X Y slices
        self.horizLine = vtk.vtkLine()
        self.vertLine = vtk.vtkLine()
        self.crosshairsActor = vtk.vtkActor()
        # self.getRenderer().AddActor(self.crosshairsActor)
        self.AddActor(self.crosshairsActor, CROSSHAIR_ACTOR)

        # rescale input image
        # contains (scale, shift)
        self.rescale = [False, (1, 0)]

        # ImageTracer
        self.imageTracer = vtk.vtkImageTracerWidget()
        # set Interactor
        self.imageTracer.SetInteractor(self.iren)
        self.imageTracer.SetCaptureRadius(1.5)
        self.imageTracer.GetLineProperty().SetColor(0.8, 0.8, 1.0)
        self.imageTracer.GetLineProperty().SetLineWidth(3.0)
        self.imageTracer.GetHandleProperty().SetColor(0.4, 0.4, 1.0)
        self.imageTracer.GetSelectedHandleProperty().SetColor(1.0, 1.0, 1.0)
        # Set the size of the glyph handle
        self.imageTracer.GetGlyphSource().SetScale(2.0)
        # Set the initial rotation of the glyph if desired.  The default glyph
        # set internally by the widget is a '+' so rotating 45 deg. gives a 'x'
        self.imageTracer.GetGlyphSource().SetRotationAngle(45.0)
        self.imageTracer.GetGlyphSource().Modified()
        self.imageTracer.ProjectToPlaneOn()
        self.imageTracer.SetHandleLeftMouseButton(True)
        # Set autoclose to on
        self.imageTracer.AutoCloseOn()
        self.imageTracer.AddObserver(vtk.vtkWidgetEvent.Select, self.style.OnTracerModifiedEvent, 1.0)

        # Slider widget
        self.sliderWidget = None
        self._sliderWidgetEnabled = enableSliderWidget

        self.__vis_mode = CILViewer2D.IMAGE_WITH_OVERLAY
        self.setVisualisationToImageWithOverlay()

    def log(self, msg):
        if self.debug:
            print(msg)

    def setInput3DData(self, imageData):
        '''alias of setInputData, kept for backward compatibility'''
        return self.setInputData(imageData)

    def setInputData(self, imageData):
        self.log("setInputData")
        self.reset()
        self.img3D = imageData
        self.installPipeline()
        self.axes_initialised = True

    def setInputData2(self, imageData):
        self.image2 = imageData
        # TODO resample on image1
        # print ("setInputData2")
        self.installPipeline2()

    def setInputAsNumpy(self,
                        numpyarray,
                        origin=(0, 0, 0),
                        spacing=(1., 1., 1.),
                        rescale=True,
                        dtype=vtk.VTK_UNSIGNED_SHORT):
        self.rescale[0] = rescale

        img = Converter.numpy2vtkImage(numpyarray, spacing, origin)

        if rescale:
            # rescale to appropriate VTK_UNSIGNED_SHORT
            stats = vtk.vtkImageAccumulate()
            stats.SetInputData(img)
            stats.Update()
            iMin = stats.GetMin()[0]
            iMax = stats.GetMax()[0]
            if (iMax - iMin == 0):
                scale = 1
            else:
                if dtype == vtk.VTK_UNSIGNED_SHORT:
                    scale = vtk.VTK_UNSIGNED_SHORT_MAX / (iMax - iMin)
                elif dtype == vtk.VTK_UNSIGNED_INT:
                    scale = vtk.VTK_UNSIGNED_INT_MAX / (iMax - iMin)

            self.rescale[1] = (scale, -iMin)
            shiftScaler = vtk.vtkImageShiftScale()
            shiftScaler.SetInputData(img)
            shiftScaler.SetScale(scale)
            shiftScaler.SetShift(-iMin)
            shiftScaler.SetOutputScalarType(dtype)
            shiftScaler.Update()
            self.img3D = shiftScaler.GetOutput()
        else:
            self.img3D = img

        self.installPipeline()

    def displaySlice(self, sliceno=[0]):
        self.setActiveSlice(sliceno)
        self.updatePipeline()
        self.renWin.Render()

    def updatePipeline(self, resetcamera=False):
        if self.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
            self.updateImageWithOverlayPipeline(resetcamera=resetcamera)
        elif self.vis_mode == CILViewer2D.RECTILINEAR_WIPE:
            self.updateRectilinearWipePipeline(resetcamera=resetcamera)

        self.AdjustCamera(resetcamera)
        self.renWin.Render()

    def updateRectilinearWipePipeline(self, resetcamera=False):
        extent = self.updateMainVOI()
        self.voi.SetVOI(*extent)
        #TODO what happens if image2 has a different sampling?
        self.voi2.SetVOI(*extent)
        self.wipeSliceMapper.SetOrientation(self.sliceOrientation)

    def updateMainVOI(self):
        # get the current slice
        extent = [i for i in self.img3D.GetExtent()]
        extent[self.sliceOrientation * 2] = self.getActiveSlice()
        extent[self.sliceOrientation * 2 + 1] = self.getActiveSlice()
        self.voi.SetVOI(extent[0], extent[1], extent[2], extent[3], extent[4], extent[5])
        self.log("extent {0}".format(extent))
        self.voi.Update()
        self.log("VOI dimensions {0}".format(self.voi.GetOutput().GetDimensions()))
        return extent

    def updateImageWithOverlayPipeline(self, resetcamera=False):
        self.updateMainVOI()
        self.ia.Update()
        self.imageSliceMapper.SetOrientation(self.sliceOrientation)
        self.imageSlice.Update()

        if self.image2 is not None:
            self.voi2.SetVOI(self.voi.GetVOI())
            self.imageSliceMapper2.SetOrientation(self.sliceOrientation)
            self.imageSlice2.Update()

        text = self.createAnnotationText("slice",
                                         (self.getActiveSlice(), self.img3D.GetDimensions()[self.sliceOrientation] - 1))
        self.updateCornerAnnotation(text, 0)

        if self.displayHistogram:
            self.updateROIHistogram()
        try:
            if not self.img3D is None:
                # print ("self.img3D" , self.img3D)
                # Set the tracer widget's projection normal to the current orientation
                self.imageTracer.SetProjectionNormal(self.sliceOrientation)
                # Set the Tracer widget's position along the current projection normal,
                # we want this on top of the image so it can be seen above all slices.
                # On the X and Z axes, slice 0 is at the top as the axis points into
                # the screen.
                slice_coords = [0, 0, 0]
                # When we are viewing the Y direction, the axis goes into the page
                # so we have to swap which end of the axis the tracer is placed so
                # that it isn't in the shadow of the slice.
                if self.getSliceOrientation() == SLICE_ORIENTATION_XZ:
                    slice_coords[self.getSliceOrientation()] = self.img3D.GetDimensions()[SLICE_ORIENTATION_XZ] - 1
                self.imageTracer.SetProjectionPosition(self.style.image2world(slice_coords)[self.getSliceOrientation()])
            else:
                print("self.img3D None")
        except Exception as ge:
            print(ge)
        self.AdjustCamera(resetcamera)
        self.renWin.Render()

    @property
    def vis_mode(self):
        return self.__vis_mode

    @vis_mode.setter
    def vis_mode(self, value):
        if value in [CILViewer2D.IMAGE_WITH_OVERLAY, CILViewer2D.RECTILINEAR_WIPE]:
            self.__vis_mode = value

    def setVisualisationPipelineMethodTo(self, method):
        if self.img3D is None:
            return
        # get the current visualisation mode
        if self.vis_mode != method:
            # remove current pipeline
            self.uninstallPipeline()
            if self.image2 is not None:
                self.uninstallPipeline2()
            # install the new pipeline
            self.vis_mode = method
            self.installPipeline()

    def setVisualisationToImageWithOverlay(self):
        self.setVisualisationPipelineMethodTo(CILViewer2D.IMAGE_WITH_OVERLAY)

    def setVisualisationToRectilinearWipe(self):
        self.setVisualisationPipelineMethodTo(CILViewer2D.RECTILINEAR_WIPE)

    def installPipeline(self):
        if self.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
            if self.img3D is not None:
                self.installImageWithOverlayPipeline()
            if self.image2 is not None:
                self.installPipeline2()
        elif self.vis_mode == CILViewer2D.RECTILINEAR_WIPE:
            self.installRectilinearWipePipeline()

        if self.getSliderWidgetEnabled():
            self.installSliceSliderWidgetPipeline()

        self.ren.ResetCamera()
        self.ren.Render()

        self.camera.SetViewUp(0, -1, 0)

        if not self.axes_initialised:
            self.camera.Azimuth(180)

        self.AdjustCamera()

        self.ren.AddViewProp(self.cursorActor)
        self.cursorActor.VisibilityOn()

        self.iren.Initialize()
        self.renWin.Render()

    def installPipeline2(self):
        if self.image2 is not None:
            if self.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
                self.installImageWithOverlayPipeline2()
            elif self.vis_mode == CILViewer2D.RECTILINEAR_WIPE:
                pass
        else:
            self.log("installPipeline2 no data")

    def installImageWithOverlayPipeline(self):
        '''Slices a 3D volume and then creates an actor to be rendered'''
        self.log("installPipeline")
        self.ren.AddViewProp(self.cornerAnnotation)

        self.voi.SetInputData(self.img3D)
        #select one slice in Z
        extent = [i for i in self.img3D.GetExtent()]
        for i in range(len(self.slicenos)):
            self.slicenos[i] = round((extent[i * 2 + 1] + extent[i * 2]) / 2)

        extent[self.sliceOrientation * 2] = self.getActiveSlice()
        extent[self.sliceOrientation * 2 + 1] = self.getActiveSlice()

        self.voi.SetVOI(extent[0], extent[1], extent[2], extent[3], extent[4], extent[5])

        self.voi.SetVOI(extent[0], extent[1], extent[2], extent[3], extent[4], extent[5])

        self.voi.Update()

        # set window/level for slice based on values in entire volume:
        self.ia.SetInputData(self.voi.GetOutput())
        self.ia.Update()
        self.style.AutoWindowLevelOnVolumeRange(update_slice=False)
        self.InitialLevel = self.getSliceColorLevel()
        self.InitialWindow = self.getSliceColorWindow()
        self.log("level {0} window {1}".format(self.InitialLevel, self.InitialWindow))

        self.imageSliceMapper.SetInputConnection(self.voi.GetOutputPort())

        if self.image_is_downsampled:
            self.imageSlice.GetProperty().SetInterpolationTypeToLinear()
        else:
            self.imageSlice.GetProperty().SetInterpolationTypeToNearest()

        self.imageSlice.Update()

        self.imageTracer.SetProjectionPosition(self.style.image2world([0, 0, 0])[self.getSliceOrientation()])

        self.AddActor(self.imageSlice, SLICE_ACTOR)

        self.imageTracer.SetViewProp(self.imageSlice)

    def installImageWithOverlayPipeline2(self):
        '''Slices a 3D volume and then creates an actor to be rendered'''
        self.log("installPipeline2")
        if self.image2 is not None:
            # render image2
            self.voi2.SetVOI(self.voi.GetVOI())
            self.voi2.SetInputData(self.image2)
            self.voi2.Update()
            lut = vtk.vtkLookupTable()

            ai = vtk.vtkImageHistogramStatistics()
            ai.SetInputConnection(self.voi2.GetOutputPort())
            ai.Update()

            self.lut2 = lut
            lut.SetNumberOfColors(7)
            lut.SetHueRange(.4, .6)
            lut.SetSaturationRange(1, 1)
            lut.SetValueRange(0.7, 0.7)
            lut.SetAlphaRange(0, 0.5)
            lut.Build()

            cov = vtk.vtkImageMapToColors()
            self.image2map = cov
            cov.SetInputConnection(self.voi2.GetOutputPort())
            cov.SetLookupTable(lut)
            cov.Update()

            self.imageSliceMapper2.SetInputConnection(cov.GetOutputPort())
            self.imageSlice2.Update()
            self.AddActor(self.imageSlice2, OVERLAY_ACTOR)
            self.ren.ResetCamera()
            self.ren.Render()

            self.AdjustCamera()

            self.iren.Initialize()
            self.renWin.Render()
        else:
            print("installPipeline2 no data")

    def installRectilinearWipePipeline(self):
        '''Create the pipeline for the rectilinear wipe'''
        self.log("installRectilinearWipePipeline")
        extent1 = list(self.img3D.GetExtent())
        #extent is slice number N
        for i in range(len(self.slicenos)):
            self.slicenos[i] = round((extent1[i * 2 + 1] + extent1[i * 2]) / 2)
        active_slice_num = self.getActiveSlice()
        orient = self.getSliceOrientation()
        extent1[orient] = active_slice_num
        extent1[orient + 1] = active_slice_num
        extent2 = extent1[:]

        self.voi.SetVOI(*extent1)
        self.voi2.SetVOI(*extent2)

        wipe = vtk.vtkImageRectilinearWipe()
        wipeSliceMapper = vtk.vtkImageSliceMapper()
        wipeSlice = vtk.vtkImageSlice()

        wipe.SetInputConnection(0, self.voi.GetOutputPort())
        wipe.SetInputConnection(1, self.voi2.GetOutputPort())
        # should set the position to center of viewport
        wipe.SetPosition(10, 10)
        wipe.SetWipe(0)

        wipeSliceMapper.SetInputConnection(wipe.GetOutputPort())
        wipeSliceMapper.SetOrientation(orient)

        wipeSlice.SetMapper(wipeSliceMapper)
        wipeSlice.GetProperty().SetInterpolationTypeToNearest()
        wipeSlice.GetProperty().SetColorLevel(self.InitialLevel)
        wipeSlice.GetProperty().SetColorWindow(self.InitialWindow)
        wipeSlice.Update()

        self.wipe = wipe
        self.wipeSliceMapper = wipeSliceMapper
        self.wipeActor = wipeSlice

        self.AddActor(wipeSlice, WIPE_ACTOR)

    def installSliceSliderWidgetPipeline(self):
        '''Create the pipeline for the slice slider widget
        
        The slider widget and representation are created if not already present.
        If present, the slider is updated.
        The slider is hidden if any of the dimensions of the visualised image is 1,
        else it is shown.
        '''
        # set slider_hidden
        dims = self.img3D.GetDimensions()
        slider_hidden = any(dim == 1 for dim in dims)

        if self.sliderWidget is not None:
            # reset the values to the appropriate ones of the new loaded image
            self.sliderCallback.update_orientation(self.style, 'reset')
            self.sliderWidget.SetEnabled(not slider_hidden)
            return

        sr = SliceSliderRepresentation()
        sr.SetMinimumValue(0)
        sr.SetMaximumValue(self.img3D.GetDimensions()[2] - 1)
        sr.SetValue(self.getActiveSlice())

        sw = vtk.vtkSliderWidget()
        sw.SetInteractor(self.getInteractor())
        sw.SetRepresentation(sr)
        sw.SetAnimationModeToAnimate()
        sw.EnabledOn()
        # enable slider
        sw.SetEnabled(not slider_hidden)
        cb = SliderCallback(self, sw)

        # Add interaction observers
        # to propagate events from the slider to the viewer
        sw.AddObserver(vtk.vtkCommand.InteractionEvent, cb)

        # propagate events from the viewer to the slider
        self.style.AddObserver("MouseWheelForwardEvent", cb.update_from_viewer, 0.9)
        self.style.AddObserver("MouseWheelBackwardEvent", cb.update_from_viewer, 0.9)
        self.style.AddObserver("CharEvent", cb.update_orientation, 0.9)

        # reset the slider
        cb.update_from_viewer(self.style, 'reset')

        # save references
        self.sliderWidget = sw
        self.sliderCallback = cb

    def uninstallSliderWidget(self):
        '''remove the slider widget from the viewer'''
        if self.sliderWidget is not None:
            sr = self.sliderWidget.GetRepresentation()
            if sr is not None:
                sr.RemoveAllObservers()
                coll = vtk.vtkPropCollection()
                sr.GetActors(coll)
                print("coll", coll)
                for actor in coll:
                    print("actor", actor)
                    self.ren.RemoveActor(actor)

    def AdjustCamera(self, resetcamera=False):
        self.ren.ResetCameraClippingRange()

        # adjust camera focal point
        camera = self.getRenderer().GetActiveCamera()
        fp = list(camera.GetFocalPoint())
        pos = list(camera.GetPosition())
        if self.flipCameraPosition:
            fp[self.sliceOrientation] = -self.getActiveSlice()
            pos[self.sliceOrientation] = -pos[self.sliceOrientation]
            self.flipCameraPosition = False
            # have to reset to false so it doesn't flip when we scroll.
        else:
            fp[self.sliceOrientation] = self.getActiveSlice()

        camera.SetFocalPoint(fp)
        camera.SetPosition(pos)

        if resetcamera:
            self.ren.ResetCamera()

    def getROI(self):
        return self.ROI

    def getROIExtent(self):
        p0 = self.ROI[0]
        p1 = self.ROI[1]
        return (p0[0], p1[0], p0[1], p1[1], p0[2], p1[2])

    ############### Handle events are moved to the interactor style

    def setSliceOrientation(self, axis):
        if axis in ['x', 'y', 'z']:
            self.getInteractor().SetKeyCode(axis)
            self.style.OnKeyPress(self.getInteractor(), "KeyPressEvent")

    def setVisualisationDownsampling(self, value):
        super().setVisualisationDownsampling(value)
        if self.image_is_downsampled:
            self.imageSlice.GetProperty().SetInterpolationTypeToLinear()
        else:
            self.imageSlice.GetProperty().SetInterpolationTypeToNearest()

    def setDisplayUnsampledCoordinates(self, value):
        self.display_unsampled_coords = value

    def getDisplayUnsampledCoordinates(self):
        return self.display_unsampled_coords

    def updateCornerAnnotation(self, text, idx=0, visibility=True):
        if visibility:
            self.cornerAnnotation.VisibilityOn()
        else:
            self.cornerAnnotation.VisibilityOff()

        self.cornerAnnotation.SetText(idx, text)
        self.iren.Render()

    def createAnnotationText(self, display_type, data):
        ''' Returns string to be set as the corner annotation, giving
        the coordinates or slice number (may be downsampled or unsampled coordinates)

        Args:
            display_type (str) :    'slice', 'pick', 'roi' - determines the
                                    way in which the data is displayed
            data (tuple):           the data to be displayed.'''

        text = None

        if isinstance(data, tuple):

            if display_type == "slice":
                if self.display_unsampled_coords and self.image_is_downsampled:
                    # Different method for converting to world coords, as we only
                    # have coordinates in one direction
                    data = list(data)
                    slice_coord = [0, 0, 0]
                    axis_length = [0, 0, 0]
                    slice_coord[self.getSliceOrientation()] = data[0]
                    axis_length[self.getSliceOrientation()] = data[1] + 1
                    slice_coord = self.style.image2world(slice_coord)[self.getSliceOrientation()]
                    axis_length = self.style.image2world(axis_length)[self.getSliceOrientation()] - 1
                    data = (round(slice_coord), round(axis_length))
                axis_label = self.axisLabelsText[self.getSliceOrientation()]
                text = "%s %d/%d" % (axis_label, data[0], data[1])

            else:
                # In all other cases, we have coordinates, and then we have an extra value
                # The extra value shouldn't be converted in the same way
                # TODO: check whether we need some kind of conversion on these 'extra' values
                if self.display_unsampled_coords and self.image_is_downsampled:
                    data = list(data)
                    extra_info = data.pop(-1)
                    data = self.style.image2world(data)
                    data.append(extra_info)
                    # Round the coordinates to integers:
                    # Without rounding the %d formatting
                    # later on just truncates to integer
                    for i, d in enumerate(data):
                        if i < 3:
                            data[i] = round(d)
                    data = tuple(data)

            if display_type == "pick":
                text = "[%d,%d,%d] : %.2g" % data

            elif display_type == "roi":
                text = "ROI: %d x %d x %d, %.2f kp" % data

        return text

    def updateROIHistogram(self):
        self.log("Updating hist")

        extent = [0 for i in range(6)]
        if self.getSliceOrientation() == SLICE_ORIENTATION_XY:
            self.log("slice orientation : XY")
            extent[0] = self.validateValue(min(self.ROI[0][0], self.ROI[1][0]), 'x')
            extent[1] = self.validateValue(max(self.ROI[0][0], self.ROI[1][0]), 'x')
            extent[2] = self.validateValue(min(self.ROI[0][1], self.ROI[1][1]), 'y')
            extent[3] = self.validateValue(max(self.ROI[0][1], self.ROI[1][1]), 'y')
            extent[4] = self.getActiveSlice()
            extent[5] = self.getActiveSlice()
            # y = abs(roi[1][1] - roi[0][1])
        elif self.getSliceOrientation() == SLICE_ORIENTATION_XZ:
            self.log("slice orientation : XZ")
            extent[0] = self.validateValue(min(self.ROI[0][0], self.ROI[1][0]), 'x')
            extent[1] = self.validateValue(max(self.ROI[0][0], self.ROI[1][0]), 'x')
            # x = abs(roi[1][0] - roi[0][0])
            extent[4] = self.validateValue(min(self.ROI[0][2], self.ROI[1][2]), 'z')
            extent[5] = self.validateValue(max(self.ROI[0][2], self.ROI[1][2]), 'z')
            # y = abs(roi[1][2] - roi[0][2])
            extent[2] = self.getActiveSlice()
            extent[3] = self.getActiveSlice()
        elif self.getSliceOrientation() == SLICE_ORIENTATION_YZ:
            self.log("slice orientation : YZ")
            extent[2] = self.validateValue(min(self.ROI[0][1], self.ROI[1][1]), 'y')
            extent[3] = self.validateValue(max(self.ROI[0][1], self.ROI[1][1]), 'y')
            # x = abs(roi[1][1] - roi[0][1])
            extent[4] = self.validateValue(min(self.ROI[0][2], self.ROI[1][2]), 'z')
            extent[5] = self.validateValue(max(self.ROI[0][2], self.ROI[1][2]), 'z')
            # y = abs(roi[1][2] - roi[0][2])
            extent[0] = self.getActiveSlice()
            extent[1] = self.getActiveSlice()

        self.log("updateROIHistogram {0}".format(extent))
        self.roiVOI.SetVOI(extent)
        self.roiVOI.SetInputData(self.img3D)
        self.roiVOI.Update()
        irange = self.roiVOI.GetOutput().GetScalarRange()

        self.roiIA.SetInputData(self.roiVOI.GetOutput())

        self.roiIA.IgnoreZeroOff()
        self.roiIA.SetComponentOrigin(int(irange[0]), 0, 0)
        self.roiIA.SetComponentSpacing(1, 0, 0)
        self.roiIA.Update()

        #use 255 bins
        delta = self.roiIA.GetMax()[0] - self.roiIA.GetMin()[0]
        nbins = 255
        self.roiIA.SetComponentSpacing(delta / nbins, 0, 0)
        self.roiIA.SetComponentExtent(0, nbins, 0, 0, 0, 0)
        self.roiIA.Update()

        self.histogramPlotActor.AddDataSetInputConnection(self.roiIA.GetOutputPort())
        self.histogramPlotActor.SetXRange(irange[0], irange[1])
        self.histogramPlotActor.SetYRange(self.roiIA.GetOutput().GetScalarRange())

    def updateLinePlot(self, imagecoordinate, display):

        self.displayLinePlot = display
        extent_x = list(self.img3D.GetExtent())
        extent_y = list(self.img3D.GetExtent())
        self.log("imagecoordinate {0}".format(imagecoordinate))

        if display:
            #extract profile along X
            if self.getSliceOrientation() == SLICE_ORIENTATION_XY:
                self.log("slice orientation : XY")
                extent_y[0] = imagecoordinate[0]
                extent_y[1] = imagecoordinate[0]

                extent_x[2] = imagecoordinate[1]
                extent_x[3] = imagecoordinate[1]

                extent_y[4] = self.getActiveSlice()
                extent_y[5] = self.getActiveSlice()
                extent_x[4] = self.getActiveSlice()
                extent_x[5] = self.getActiveSlice()

                self.linePlotActor.SetDataObjectXComponent(0, 0)
                self.linePlotActor.SetDataObjectXComponent(1, 1)

                #y = abs(roi[1][1] - roi[0][1])
            elif self.getSliceOrientation() == SLICE_ORIENTATION_XZ:
                self.log("slice orientation : XZ")
                extent_y[0] = imagecoordinate[0]
                extent_y[1] = imagecoordinate[0]
                #x = abs(roi[1][0] - roi[0][0])
                extent_x[4] = imagecoordinate[2]
                extent_x[5] = imagecoordinate[2]
                #y = abs(roi[1][2] - roi[0][2])
                extent_x[2] = self.getActiveSlice()
                extent_x[3] = self.getActiveSlice()
                extent_y[2] = self.getActiveSlice()
                extent_y[3] = self.getActiveSlice()
                self.linePlotActor.SetDataObjectXComponent(0, 0)
                self.linePlotActor.SetDataObjectXComponent(1, 2)

            elif self.getSliceOrientation() == SLICE_ORIENTATION_YZ:
                self.log("slice orientation : YZ")
                extent_y[2] = imagecoordinate[1]
                extent_y[3] = imagecoordinate[1]
                #x = abs(roi[1][1] - roi[0][1])
                extent_x[4] = imagecoordinate[2]
                extent_x[5] = imagecoordinate[2]
                #y = abs(roi[1][2] - roi[0][2])
                extent_x[0] = self.getActiveSlice()
                extent_x[1] = self.getActiveSlice()
                extent_y[0] = self.getActiveSlice()
                extent_y[1] = self.getActiveSlice()

                self.linePlotActor.SetDataObjectXComponent(0, 1)
                self.linePlotActor.SetDataObjectXComponent(1, 2)

            self.log("x {0} extent_x {1}".format(imagecoordinate[0], extent_x))
            self.log("y {0} extent_y {1}".format(imagecoordinate[1], extent_y))
            self.lineVOIX.SetVOI(extent_x)
            self.lineVOIX.SetInputData(self.img3D)
            self.lineVOIX.Update()
            self.lineVOIY.SetVOI(extent_y)
            self.lineVOIY.SetInputData(self.img3D)
            self.lineVOIY.Update()

            # fill
            if self.linePlot == 0:
                self.linePlotActor.AddDataSetInputConnection(self.lineVOIX.GetOutputPort())
                self.linePlotActor.AddDataSetInputConnection(self.lineVOIY.GetOutputPort())
                # self.getRenderer().AddActor(self.linePlotActor)
                self.AddActor(self.linePlotActor, LINEPLOT_ACTOR)
                self.linePlot = 1

            # Set position of line plot

            # Set up character sizes and borders
            width = 8
            border = 12
            height = 12

            # Calculate the world origin in display coordinates
            origin_display = self.style.world2display((0, 0, 0))

            # Calculate the offset due to labels and borders on the graphs y-axis
            max_y = max(int(self.lineVOIY.GetOutput().GetScalarRange()[1]),
                        int(self.lineVOIX.GetOutput().GetScalarRange()[1]))
            y_digits = len(str(max_y))
            x_min_offset = (width * y_digits) + border
            y_min_offset = height + border

            # Offset the origin to account for the labels
            origin_nview = self.style.display2normalisedViewport(
                (origin_display[0] - x_min_offset, origin_display[1] - y_min_offset))

            # Calculate the far right border
            data_extent = self.style.GetDataExtentInWorld()
            top_right_world = self.style.GetMinMaxVoxelsFromExtent(data_extent)[1]
            top_right_disp = self.style.world2display(top_right_world)
            top_right_nview = self.style.display2normalisedViewport(
                (top_right_disp[0] + border + height, top_right_disp[1]))

            # Set the position on the image
            self.linePlotActor.SetPosition(origin_nview)
            self.linePlotActor.SetPosition2(top_right_nview[0] - origin_nview[0], 0.4)

            self.log("data length x {0} y {1}".format(self.lineVOIX.GetOutput().GetDimensions(),
                                                      self.lineVOIY.GetOutput().GetDimensions()))
            self.linePlotActor.VisibilityOn()
            self.crosshairsActor.VisibilityOn()

            self.renWin.Render()

        else:
            self.linePlotActor.VisibilityOff()
            self.crosshairsActor.VisibilityOff()

            self.renWin.Render()

    def AddActor(self, actor, name=None):
        '''print("ADDING ACTOR", name)
        self.log("Calling AddActor " + name)
        present_actors = self.ren.GetActors() # Only seems to return some of the actors - possibly only the visible ones?
        present_actors.InitTraversal()
        self.log("Currently present actors {}".format(present_actors))

        print("Current len", present_actors.GetNumberOfItems())
    
        for i in range(present_actors.GetNumberOfItems()):
            nextActor = present_actors.GetNextActor()
            nextActor.SetVisibility(False)
            self.log("{} {} Visibility {}".format(i, nextActor, nextActor.GetVisibility() ))
            self.log("ClassName"+ str( nextActor.GetClassName()))

        
        print("intermediate len", self.ren.GetActors().GetNumberOfItems())        
        if name is None:
            name = 'actor_{}'.format(present_actors.GetNumberOfItems()+1)'''

        self.ren.AddActor(actor)
        self.actors[name] = actor

    def GetActorsDict(self):
        return self.actors

    def GetActor(self, name):
        if name in self.actors:
            actor = self.actors[name]
            return actor
        else:
            return (None)

    def uninstallPipeline(self):
        '''uninstall the current visualisation pipeline'''
        if self.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
            self.removeActor([SLICE_ACTOR, HISTOGRAM_ACTOR])
        elif self.vis_mode == CILViewer2D.RECTILINEAR_WIPE:
            self.removeActor(WIPE_ACTOR)

    def removeActor(self, actor):
        '''remove named actor'''
        if not isinstance(actor, (list, tuple)):
            actor = [actor]
        keys = self.actors.keys()
        for a in actor:
            if a in keys:
                self.ren.RemoveActor(self.actors[a])
                self.actors.pop(a)

    def uninstallPipeline2(self):
        if self.vis_mode == CILViewer2D.IMAGE_WITH_OVERLAY:
            self.removeActor(OVERLAY_ACTOR)
        elif self.vis_mode == CILViewer2D.RECTILINEAR_WIPE:
            # rectilinear wipe visualises 2 images in the same pipeline
            pass

    def reset(self):
        self.uninstallPipeline()
        if self.image2 is not None:
            self.uninstallPipeline2()
        if self.getSliderWidgetEnabled():
            self.uninstallSliderWidget()

    def getSliderWidgetEnabled(self):
        return self._sliderWidgetEnabled

    def setSliderWidgetEnabled(self, enable):
        if enable:
            self._sliderWidgetEnabled = enable



import os
import sys
from functools import partial
from pathlib import Path
import ccpi.viewer.viewerLinker as vlink
import vtk
from ccpi.viewer import CILViewer2D, CILViewer
from ccpi.viewer.CILViewer2D import CILViewer2D
from ccpi.viewer.CILViewer import CILViewer
from ccpi.viewer.QCILViewerWidget import QCILDockableWidget
from ccpi.viewer.ui.dialogs import HDF5InputDialog, RawInputDialog, ViewerSettingsDialog
from ccpi.viewer.ui.qt_widgets import ViewerCoordsDockWidget
from ccpi.viewer.utils import cilPlaneClipper
from ccpi.viewer.utils.io import ImageReader
from eqt.threading import Worker
from eqt.ui.SessionDialogs import ErrorDialog
from eqt.ui.MainWindowWithSessionManagement import MainWindowWithProgressDialogs, MainWindowWithSessionManagement
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QApplication, QFileDialog, QMainWindow, QSizePolicy


class ViewerMainWindow(MainWindowWithProgressDialogs):
    ''' Creates a window which is designed to house one or more viewers.
    Note: does not create a viewer as we don't know whether the user would like it to exist in a
    dockwidget or central widget
    Instead it creates a main window with many of the tools needed for a window
    that will house a viewer. 
    
    Methods of Interest
    -------------------
    self.createViewerCoordsDockWidget
        Creates a dock widget that displays the shape of the loaded image
        and if downsampled, the shape of the downsampled image. Allows user
        to select whether the coordinates on the viewer are shown in the
        original or downsampled image system.
    
    Assumes that at least one viewer is present, saved in self.viewers

    Parameters
    ----------
    title : str, optional
        The title of the window. The default is "ViewerMainWindow".
    app_name : str, optional
        The name of the application. The default is "ViewerMainWindow".
    settings_name : str, optional
        The name of the settings file. The default is None.
        If None, this is set to the app name.
    organisation_name : str, optional
        The name of the organisation. The default is None.
        If None, this is set to the app name.
    '''

    def __init__(self,
                 title="ViewerMainWindow",
                 app_name="ViewerMainWindow",
                 settings_name=None,
                 organisation_name=None):

        super(ViewerMainWindow, self).__init__(title,
                                               app_name,
                                               settings_name=settings_name,
                                               organisation_name=organisation_name)

        self.default_downsampled_size = 512**3

        self._viewers = []
        self._viewer_docks = []
        self._frames = []
        self._vs_dialog = None

        self.createViewerCoordsDockWidget()

    @property
    def viewers(self):
        '''The list of viewers in the window'''
        return self._viewers

    @property
    def viewer_docks(self):
        '''The list of dock widgets that contain the viewers'''
        return self._viewer_docks

    @property
    def frames(self):
        '''The list of frames in the window'''
        return self._frames

    @property
    def vs_dialog(self):
        '''The viewer settings dialog'''
        return self._vs_dialog

    @vs_dialog.setter
    def vs_dialog(self, value):
        self._vs_dialog = value

    def createAppSettingsDialog(self):
        '''Create a dialog to change the application settings, or open an existing one.
        This is a method in the MainWindowWithSessionManagement class, which we
        override here to make our own settings dialog'''
        if self.vs_dialog is None:
            self.vs_dialog = ViewerSettingsDialog(self)
            self.vs_dialog.Ok.clicked.connect(lambda: self.onAppSettingsDialogAccepted(self.vs_dialog))
            self.vs_dialog.Ok.clicked.connect(self.acceptViewerSettings)
            self.vs_dialog.Cancel.clicked.connect(self.vs_dialog.close)
        self.setAppSettingsDialogWidgets(self.vs_dialog)
        self.setViewerSettingsDialogWidgets()
        self.vs_dialog.open()

    def setViewerSettingsDialogWidgets(self):
        '''Set the viewer-specific widgets on the app settings dialog, based on the 
        current settings of the app.
        '''
        sw = self.vs_dialog.widgets

        if self.settings.value("vis_size") is not None:
            sw['vis_size_field'].setValue(float(self.settings.value("vis_size")))
        else:
            sw['vis_size_field'].setValue(self.getDefaultDownsampledSize() / (1024**3))

        if self.settings.value("use_gpu_volume_mapper") is not None:
            use_gpu = str(self.settings.value("use_gpu_volume_mapper"))
            sw['gpu_checkbox_field'].setChecked(use_gpu == "true")
        else:
            sw['gpu_checkbox_field'].setChecked(True)

    def acceptViewerSettings(self):
        '''This is called when the user clicks the OK button on the
        app settings dialog.
        Saves the viewer settings to the QSettings object.
        '''

        self.settings.setValue("vis_size", float(self.vs_dialog.widgets['vis_size_field'].value()))

        # Check if the user has changed the volume mapper setting:
        current_setting = self.settings.value("use_gpu_volume_mapper")

        if current_setting == str(self.vs_dialog.widgets['gpu_checkbox_field'].isChecked()):
            return
        else:
            bool_to_volume_mapper = {True: vtk.vtkSmartVolumeMapper(), False: vtk.vtkFixedPointVolumeRayCastMapper()}

            use_gpu_volume_mapper = self.vs_dialog.widgets['gpu_checkbox_field'].isChecked()

            self.settings.setValue("use_gpu_volume_mapper", use_gpu_volume_mapper)

            for viewer in self.viewers:
                if isinstance(viewer, CILViewer):
                    viewer.setVolumeMapper(bool_to_volume_mapper[use_gpu_volume_mapper])
                    if viewer.volume_render_initialised:
                        viewer.installVolumeRenderActorPipeline()
                        viewer.updatePipeline()

    def selectImage(self, label=None):
        ''' This selects opens a file dialog for the user to select the image
        and gets the user to enter relevant data, but does not load them on a viewer yet.
        
        Parameters
        ----------
        label : QLabel, optional
            The label to display the basename of the file name on.
        
        Returns
        -------
        file : str
            The file name of the image selected by the user.
        '''
        dialog = QFileDialog()
        files = list(dialog.getOpenFileNames(self, "Select Images")[0])
        self.raw_attrs = {}
        self.hdf5_attrs = {}
        if files is None:
            return
        if len(files) == 1:
            file = files[0]
            file_extension = Path(file).suffix.lower()
            if 'raw' in file_extension:
                if hasattr(self, 'raw_dialog'):
                    self.raw_dialog.restoreAllSavedWidgetStates()
                else:
                    raw_dialog = RawInputDialog(self, file)
                    raw_dialog.Ok.clicked.connect(lambda: self.getRawAttrsFromDialog(raw_dialog))
                    self.raw_dialog = raw_dialog
                # See https://doc.qt.io/qt-6/qdialog.html#exec
                # Shows a modal dialog, blocking until the user closes it.
                raw_dialog.exec()
                if self.raw_attrs == {}:
                    return None
            elif file_extension in ['.nxs', '.h5', '.hdf5']:
                dialog = HDF5InputDialog(self, file)
                dialog.Ok.clicked.connect(lambda: self.getHDF5AttrsFromDialog(dialog))
                # See https://doc.qt.io/qt-6/qdialog.html#exec
                # Shows a modal dialog, blocking until the user closes it.
                dialog.exec()
                if self.hdf5_attrs == {}:
                    return None
            self.input_dataset_file = file
        else:
            self.input_dataset_file = files

        if label is not None:
            label.setText(os.path.basename(files[0]))

        return self.input_dataset_file

    def getRawAttrsFromDialog(self, dialog):
        '''Gets the raw attributes from the dialog and saves them to the
        raw_attrs attribute of the window.
        Also closes the dialog.
        
        Parameters
        ----------
        dialog : RawInputDialog
            The dialog to get the attributes from.
        '''
        dialog.saveAllWidgetStates()
        self.raw_attrs = dialog.getRawAttrs()
        dialog.close()

    def getHDF5AttrsFromDialog(self, dialog):
        '''
        Gets the HDF5 attributes from the dialog and saves them to the
        hdf5_attrs attribute of the window.
        Also closes the dialog.
        
        Parameters
        ----------
        dialog : HDF5InputDialog
            The dialog to get the attributes from.
        '''

        self.hdf5_attrs = dialog.getHDF5Attributes()
        dialog.close()

    def createViewerCoordsDockWidget(self):
        '''
        Creates a dock widget which contains widgets for displaying the 
        image shown on the viewer, and the coordinate system of the viewer.
        '''
        dock = ViewerCoordsDockWidget(self)
        self.viewer_coords_dock = dock
        self.viewer_coords_dock.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        dock.getWidgets()['coords_combo_field'].currentIndexChanged.connect(self.updateViewerCoords)

    def placeViewerCoordsDockWidget(self):
        '''
        Places the viewer coords dock widget in the main window.
        May be overridden by subclasses to place the dock widget in a different
        location.
        This method is a courtesy method and gets never explicitly called.
        So it's the user's responsibility to place the widget, for example by
        calling this method.
        '''
        self.addDockWidget(Qt.BottomDockWidgetArea, self.viewer_coords_dock)

    def setViewersInputFromDialog(self, viewers, input_num=1):
        '''Displays an image in the viewer(s).'''
        image_file = self.selectImage()
        if image_file is not None:
            self.setViewersInput(image_file, viewers, input_num=input_num)

    def setViewersInput(self, image, viewers, input_num=1, image_name=None):
        '''
        Displays an image in the viewer/s.

        Parameters
        ----------
        viewer: CILViewer2D or CILViewer, or list of CILViewer2D or CILViewer
            The viewer(s) to display the image in.
        image: str, list or vtk.vtkImageData
            The image or images to display. If a string, it is assumed to be a file name.
        input_num : int
            The input number to the viewer. 1 or 2, where 1 is the default image and 2 is the overlay
            image. Only used if the viewer is a 2D viewer.
        image_name : str
            The name of the image. If not given, the file name is used.
            Must be set if no file name is given.
        '''
        raw_image_attrs = self.raw_attrs
        hdf5_image_attrs = self.hdf5_attrs
        dataset_name = hdf5_image_attrs.get('dataset_name')
        resample_z = hdf5_image_attrs.get('resample_z', True)
        target_size = self.getTargetImageSize()
        if isinstance(image, str) or isinstance(image, list):
            image_reader = ImageReader(file_name=image)
        else:
            image_reader = ImageReader(vtk_image=image)
        image_reader.SetTargetSize(target_size)
        image_reader.SetRawImageAttributes(raw_image_attrs)
        image_reader.SetHDF5DatasetName(dataset_name)
        image_reader.SetResampleZ(resample_z)
        image_reader_worker = Worker(image_reader.Read)
        self.createUnknownProgressWindow("Reading Image")
        if image_name is None and isinstance(image, str):
            image_name = image
        image_reader_worker.signals.result.connect(
            partial(self.displayImage, viewers, input_num, image_reader, image_name))
        image_reader_worker.signals.finished.connect(lambda: self.finishProcess("Reading Image"))
        image_reader_worker.signals.error.connect(self.processErrorDialog)
        self.threadpool.start(image_reader_worker)

    def processErrorDialog(self, error, **kwargs):
        '''
        Creates an error dialog to display the error.
        '''
        dialog = ErrorDialog(self, "Error", str(error[1]), str(error[2]))
        dialog.open()

    def displayImage(self, viewers, input_num, reader, image_name, image):
        '''
        Displays an image on the viewer/s.

        Parameters
        ----------
        viewers: CILViewer2D or CILViewer, or list of CILViewer2D or CILViewer
            The viewer(s) to display the image on.
        input_num : int
            The input number to the viewer. 1 or 2. Only used if the viewer is
            a 2D viewer. 1 is the default image and 2 is the overlay image.
        reader: ImageReader
            The reader used to read the image. This contains some extra info about the
            original image file.
        image_name: str
            The image file name, or label to display on the viewer.
        image: vtkImageData
            The image to display.

        '''
        if image is None:
            return
        if not isinstance(viewers, list):
            viewers = [viewers]
        for viewer in viewers:
            if input_num == 1:
                viewer.setInputData(image)
            elif input_num == 2:
                if isinstance(viewer, CILViewer2D):
                    viewer.setInputData2(image)
                # If input_num=2 we don't want to update the viewer coords dock widget
                # with the name of the image file, as this is the overlay image.
                image_name = None
        self.updateGUIForNewImage(reader, viewer, image_name)

    def updateGUIForNewImage(self, reader=None, viewer=None, image_name=None):
        '''
        Updates the GUI for a new image:
        - Updates the viewer coordinates dock widget with the displayed image dimensions
            and the loaded image dimensions (if a reader is provided)
        - Updates the viewer coordinates dock widget with the image file name
        In subclass, may want to add more functionality.
        '''
        self.updateViewerCoordsDockWidgetWithCoords(reader)
        self.updateViewerCoordsDockWidgetWithImageFileName(image_name)

    def updateViewerCoordsDockWidgetWithImageFileName(self, image_name=None):
        '''
        Updates the viewer coordinates dock widget with the image file name.

        Parameters
        ----------
        image_file: str
            The image file name.
        '''
        if image_name is None:
            return

        widgets = self.viewer_coords_dock.getWidgets()
        widgets['image_field'].clear()
        widgets['image_field'].addItem(image_name)
        widgets['image_field'].setCurrentIndex(0)

    def updateViewerCoordsDockWidgetWithCoords(self, reader=None):
        '''
        Updates the viewer coordinates dock widget with the displayed image dimensions
        and the loaded image dimensions (if a reader is provided)
        
        Parameters
        ----------
        reader: ImageReader
            The reader used to read the image. This contains some extra info about the
            original image file.
        '''
        viewer = self.viewer_coords_dock.viewers[0]

        if not isinstance(viewer, (CILViewer2D, CILViewer)):
            return

        image = viewer.img3D

        if image is None:
            return

        widgets = self.viewer_coords_dock.getWidgets()

        # Update the coordinates: ------------------------------

        displayed_image_dims = str(list(image.GetDimensions()))

        widgets['coords_combo_field'].setCurrentIndex(0)

        widgets['loaded_image_dims_field'].setVisible(True)
        widgets['loaded_image_dims_label'].setVisible(True)

        if reader is None:
            # If reader is None, then we have no info about the original
            # image file, or whether it was resampled.
            resampled = False
        else:
            loaded_image_attrs = reader.GetLoadedImageAttrs()
            resampled = loaded_image_attrs.get('resampled')

            if resampled:
                # Check if the image was actually resampled:
                original_image_attrs = reader.GetOriginalImageAttrs()
                original_image_dims = original_image_attrs.get('shape')
                if list(original_image_dims) == list(displayed_image_dims):
                    resampled = False

        widgets['coords_warning_field'].setVisible(resampled)
        widgets['coords_info_field'].setVisible(resampled)
        widgets['coords_combo_field'].setEnabled(resampled)
        widgets['displayed_image_dims_field'].setVisible(resampled)
        widgets['displayed_image_dims_label'].setVisible(resampled)

        if not resampled:
            widgets['loaded_image_dims_field'].setText(displayed_image_dims)
        else:
            original_image_attrs = reader.GetOriginalImageAttrs()
            original_image_dims = str(list(original_image_attrs.get('shape')))
            spacing = image.GetSpacing()
            for _viewer in self.viewer_coords_dock.viewers:
                _viewer.setVisualisationDownsampling(spacing)

            widgets['loaded_image_dims_field'].setText(original_image_dims)
            widgets['displayed_image_dims_field'].setText(displayed_image_dims)

    def updateViewerCoords(self):
        '''
        Updates the coordinate system of the viewer/s associated with the 
        ViewerCoordinatesDockWidget and the displayed image
        dimensions, for the image on that viewer.

        '''
        viewer_coords_widgets = self.viewer_coords_dock.getWidgets()
        viewer = self.viewer_coords_dock.viewers[0]
        shown_resample_rate = viewer.getVisualisationDownsampling()
        for viewer in self.viewer_coords_dock.viewers:
            if isinstance(viewer, CILViewer2D) and viewer.img3D is not None:
                if viewer_coords_widgets['coords_combo_field'].currentIndex() == 0:
                    viewer.setDisplayUnsampledCoordinates(True)
                    if shown_resample_rate != [1, 1, 1]:
                        viewer_coords_widgets['coords_warning_field'].setVisible(True)
                    else:
                        viewer_coords_widgets['coords_warning_field'].setVisible(False)
                else:
                    viewer.setDisplayUnsampledCoordinates(False)
                    viewer_coords_widgets['coords_warning_field'].setVisible(False)

                viewer.updatePipeline()

    def getTargetImageSize(self):
        ''' Get the target size for an image to be displayed in bytes.'''
        if self.settings.value("vis_size") is not None:
            target_size = float(self.settings.value("vis_size")) * (1024**3)
        else:
            target_size = self.getDefaultDownsampledSize()
        return target_size

    def setDefaultDownsampledSize(self, value):
        ''' Set the default size for an image to be displayed in bytes'''
        self.default_downsampled_size = int(value)

    def getDefaultDownsampledSize(self):
        ''' Get the default size for an image to be displayed in bytes'''
        return self.default_downsampled_size


class ViewerMainWindowWithSessionManagement(MainWindowWithSessionManagement, ViewerMainWindow):
    ''' Creates a window which is designed to house one or more viewers.
    Note: does not create a viewer as we don't know whether the user would like it to exist in a
    dockwidget or central widget
    Instead it creates a main window with many of the tools needed for a window
    that will house a viewer.

    Parameters
    ----------
    title : str, optional
        The title of the window. The default is "ViewerMainWindowWithSessionManagement".
    app_name : str, optional
        The name of the application. The default is "ViewerMainWindowWithSessionManagement".
    settings_name : str, optional
        The name of the settings file. The default is None.
        If None, this is set to the app name.
    organisation_name : str, optional
        The name of the organisation. The default is None.
        If None, this is set to the app name.
    '''

    def __init__(self,
                 title="ViewerMainWindowWithSessionManagement",
                 app_name="ViewerMainWindowWithSessionManagement",
                 settings_name=None,
                 organisation_name=None):

        super(ViewerMainWindowWithSessionManagement, self).__init__(title,
                                                                    app_name,
                                                                    settings_name=settings_name,
                                                                    organisation_name=organisation_name)


class TwoViewersMainWindowMixin(object):
    '''
    Provides a mixin for a TwoViewersMainWindow or a TwoViewersMainWindowWithSessionManagement class.
    Provides the setupTwoViewers method, which:
    creates a window containing two viewers, both in dockwidgets.
    The viewers are linked together, so that they share the same
    camera position and orientation.

    Also provides other methods for adding viewers, and for
    setting up the viewer coordinates dockwidget.
    
    If a viewer is 2D, then a :code:`cilPlaneClipper` will be added to the viewer
    to clip any 3D object in the view, such as a sphere source.
    
    Properties of note:
        viewers: a list of the two viewers
        frames: a list of the two frames
        viewer_docks: a list of the two viewer docks


    '''

    def setupTwoViewers(self, viewer1_type, viewer2_type=None):
        '''
        Adds the viewers as dockwidgets inside the central widget.
        Also adds a ViewerCoordinatesDockWidget to the window.

        Parameters
        ----------
        viewer1_type: '2D' or '3D'
            The type of the first viewer.
            If '2D', then a CILViewer2D is created.
            If '3D', then a CILViewer is created.
        viewer2_type: '2D' or '3D', or None, optional
            The type of the second viewer.
            If '2D', then a CILViewer2D is created.
            If '3D', then a CILViewer is created.
            If None, then no second viewer is created.
        '''
        if not hasattr(self, 'central_widget'):
            cw = QMainWindow()
            cw.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
            self.setCentralWidget(cw)
            self.central_widget = cw

        self.addViewer(viewer1_type)

        if viewer2_type is not None:
            self.addViewer(viewer2_type)

        self.placeViewerCoordsDockWidget()

        self.viewer_coords_dock.viewers = self.viewers

        self.setupPlaneClipping()

    def addViewer(self, viewer_type):
        '''
        Add a viewer to the window, inside a DockWidget, within the
        central widget.

        Saves the viewer to self.viewers, saves the frame to self.frames,
        and saves the dock to self.viewer_docks.

        If any viewers are already present, the new viewer will be added
        to the right of the window.
        
        Parameters
        ----------
        viewer_type: '2D' or '3D'
            The type of viewer to add to the window.
        '''
        if len(self.viewers) == 2:
            raise ValueError("Cannot add more than two viewers to this window.")

        if viewer_type == '2D':
            interactor_style = vlink.Linked2DInteractorStyle
            dock_title = "2D View"
            viewer = CILViewer2D
        elif viewer_type == '3D':
            interactor_style = vlink.Linked3DInteractorStyle
            dock_title = "3D View"
            viewer = CILViewer
        else:
            raise ValueError("viewer_type must be either '2D' or '3D'")

        dock = QCILDockableWidget(viewer=viewer, shape=(600, 600), interactorStyle=interactor_style, title=dock_title)
        dock.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        if len(self.viewers) == 0:
            self.central_widget.addDockWidget(Qt.LeftDockWidgetArea, dock)
        elif len(self.viewers) == 1:
            self.central_widget.addDockWidget(Qt.RightDockWidgetArea, dock)
            self.linkedViewersSetUp([self.viewers[0], dock.frame.viewer])
            self.linker.enable()

        self.viewers.append(dock.frame.viewer)
        self.frames.append(dock.frame)
        self.viewer_docks.append(dock)

    def linkedViewersSetUp(self, viewers):
        '''
        Link two viewers together.
        
        Parameters
        ----------
        viewers : list of CILViewer or CILViewer2D
            The viewers to link together.
        '''
        v1 = viewers[0]
        v2 = viewers[1]
        self.linker = vlink.ViewerLinker(v1, v2)
        self.linker.setLinkPan(False)
        self.linker.setLinkZoom(False)
        self.linker.setLinkWindowLevel(True)
        self.linker.setLinkSlice(True)

    def showHideViewer(self, viewer_num, show=True):
        '''Show or  hide a viewer.
        
        Parameters
        ----------
        viewer_num : int
            The index of the viewer to show or hide, in the list self.viewers.
        show : bool, optional
            Whether to show the viewer. Default is True.
        '''
        if not show:
            self.viewer_docks[viewer_num].hide()
        else:
            self.viewer_docks[viewer_num].show()

    def setupPlaneClipping(self):
        '''
        Set up plane clipping for all 2D viewers.
        '''
        for viewer in self.viewers:
            if isinstance(viewer, CILViewer2D):
                viewer.PlaneClipper = cilPlaneClipper()
                viewer.PlaneClipper.SetInteractorStyle(viewer.style)
                viewer.style.AddObserver("MouseWheelForwardEvent", viewer.PlaneClipper.UpdateClippingPlanes, 0.9)
                viewer.style.AddObserver("MouseWheelBackwardEvent", viewer.PlaneClipper.UpdateClippingPlanes, 0.9)
                viewer.style.AddObserver("KeyPressEvent", viewer.PlaneClipper.UpdateClippingPlanes, 0.9)

    def placeViewerCoordsDockWidget(self):
        '''
        Positions the viewer coords dock widget below the viewers
        '''
        self.addDockWidget(Qt.BottomDockWidgetArea, self.viewer_coords_dock)


class TwoViewersMainWindow(TwoViewersMainWindowMixin, ViewerMainWindow):
    '''
    Creates a window containing two viewers, both in dockwidgets.
    The viewers are linked together, so that they share the same
    camera position and orientation.
    
    If a viewer is 2D, then a :code:`cilPlaneClipper` will be added to the
    viewer to clip any 3D object in the view, such as a sphere source.
    
    Properties of note:
        viewers: a list of the two viewers
        frames: a list of the two frames
        viewer_docks: a list of the two viewer docks
    
    Parameters
    ----------
    title : str, optional
        The title of the window. The default is "TwoViewersMainWindow".
    app_name : str, optional
        The name of the application. The default is "TwoViewersMainWindow".
    settings_name : str, optional
        The name of the settings file. The default is None.
        If None, this is set to the `app_name`.
    organisation_name : str, optional
        The name of the organisation. The default is None.
        If None, this is set to the `app_name`.
    viewer1_type: '2D' or '3D', default '2D'
        The type of the first viewer. 
        If '2D', then a CILViewer2D is created.
        If '3D', then a CILViewer is created.
    viewer2_type: '2D' or '3D', or None, optional. Default '3D'
        The type of the second viewer. 
        If '2D', then a CILViewer2D is created.
        If '3D', then a CILViewer is created.
        If None, then no second viewer is created.
    '''

    def __init__(self,
                 title="TwoViewersMainWindow",
                 app_name="TwoViewersMainWindow",
                 settings_name=None,
                 organisation_name=None,
                 viewer1_type='2D',
                 viewer2_type='3D'):

        super(TwoViewersMainWindow, self).__init__(title, app_name, settings_name, organisation_name)
        self.setupTwoViewers(viewer1_type, viewer2_type)


class TwoViewersMainWindowWithSessionManagement(TwoViewersMainWindowMixin, ViewerMainWindowWithSessionManagement):
    '''
    Creates a window containing two viewers, both in dockwidgets.
    This main window has methods for saving and loading sessions.

    The viewers are linked together, so that they share the same
    camera position and orientation.
    
    If a viewer is 2D, then a :code:`cilPlaneClipper` will be added to the
    viewer to clip any 3D object in the view, such as a sphere source.
    
    Properties of note:
        viewers: a list of the two viewers
        frames: a list of the two frames
        viewer_docks: a list of the two viewer docks

    This class is meant to be subclassed, and the subclass should implement the following methods:
     - getSessionConfig
     - finishLoadConfig
     as these deal with the session saving and loading. See 'MainWindowWithSessionManagement' for more details.

    
    Parameters
    ----------
    title : str, optional
        The title of the window.
        The default is "TwoViewersMainWindowWithSessionManagement".
    app_name : str, optional
        The name of the application.
        The default is "TwoViewersMainWindowWithSessionManagement".
    settings_name : str, optional
        The name of the settings file. The default is None.
        If None, this is set to the app name.
    organisation_name : str, optional
        The name of the organisation. The default is None.
        If None, this is set to the app name.
    viewer1_type: '2D' or '3D', default '2D'
        The type of the first viewer. 
        If '2D', then a CILViewer2D is created.
        If '3D', then a CILViewer is created.
    viewer2_type: '2D' or '3D', or None, optional. Default '3D'
        The type of the second viewer. 
        If '2D', then a CILViewer2D is created.
        If '3D', then a CILViewer is created.
        If None, then no second viewer is created.
    '''

    def __init__(self,
                 title="TwoViewersMainWindowWithSessionManagement",
                 app_name="TwoViewersMainWindowWithSessionManagement",
                 settings_name=None,
                 organisation_name=None,
                 viewer1_type='2D',
                 viewer2_type='3D'):
        super(TwoViewersMainWindowWithSessionManagement, self).__init__(title, app_name, settings_name,
                                                                        organisation_name)
        self.setupTwoViewers(viewer1_type, viewer2_type)


# For running example window -----------------------------------------------------------


def create_main_window():
    window = TwoViewersMainWindow("Test Viewer Main Window", 'Test Viewer Main Window')
    return window


def main():
    app = QApplication(sys.argv)
    window = create_main_window()
    window.show()
    app.exec_()


if __name__ == "__main__":
    main()
